import unittest
from unittest import mock

from lgrez import config, bdd
from lgrez.features import voter_agir
from test import mock_discord, mock_bdd, mock_env


class TestVoterAgirFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.voter_agir utility functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db      # Empty database for this method
    @mock_env.patch_env(LGREZ_DATA_SHEET_ID="uiz")
    @mock.patch("lgrez.blocs.gsheets.connect")
    async def test_export_vote(self, gconnect_patch):
        """Unit tests for voter_agir.export_vote function."""
        # def export_vote(vote, joueur)
        export_vote = voter_agir.export_vote
        mock_bdd.add_campsroles(10, 10)
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            _role_slug="role7", _camp_slug="camp8",
                            vote_condamne_="oh", vote_maire_="ah",
                            vote_loups_="eh")
        joueur.add()
        bdd.BaseAction(slug="ouiZ").add()
        bdd.BaseAction(slug="nonZ").add()
        bdd.BaseAction(slug="lalaZ").add()
        bdd.Action(joueur=joueur, _base_slug="ouiZ", decision_="dZ1").add()
        bdd.Action(joueur=joueur, _base_slug="nonZ", decision_="dZ2").add()
        bdd.Action(joueur=joueur, _base_slug="lalaZ", decision_=None).add()

        # no LGREZ_DATA_SHEET_ID
        with mock_env.patch_env(LGREZ_DATA_SHEET_ID=None):
            with self.assertRaises(RuntimeError):
                export_vote("cond", joueur)
        gconnect_patch.assert_not_called()

        # vote = bad value
        with self.assertRaises(ValueError):
            export_vote("bzz", joueur)
        gconnect_patch.assert_not_called()

        # vote = "cond"
        export_vote("cond", joueur)
        gconnect_patch.assert_called_once_with("uiz")
        gconnect_patch().worksheet.assert_called_with("votecond_brut")
        gconnect_patch().worksheet().append_row.assert_called_once()
        appened = gconnect_patch().worksheet().append_row.call_args.args[0]
        self.assertEqual(["Joueur1", "oh"], appened[1:])
        gconnect_patch.reset_mock(return_value=True)

        # vote = "maire"
        export_vote("maire", joueur)
        gconnect_patch.assert_called_once_with("uiz")
        gconnect_patch().worksheet.assert_called_with("votemaire_brut")
        gconnect_patch().worksheet().append_row.assert_called_once()
        appened = gconnect_patch().worksheet().append_row.call_args.args[0]
        self.assertEqual(["Joueur1", "ah"], appened[1:])
        gconnect_patch.reset_mock(return_value=True)

        # vote = "loups"
        export_vote("loups", joueur)
        gconnect_patch.assert_called_once_with("uiz")
        gconnect_patch().worksheet.assert_called_with("voteloups_brut")
        gconnect_patch().worksheet().append_row.assert_called_once()
        appened = gconnect_patch().worksheet().append_row.call_args.args[0]
        self.assertEqual(["Joueur1", "camp8", "eh"], appened[1:])
        gconnect_patch.reset_mock(return_value=True)

        # vote = "action"
        export_vote("action", joueur)
        gconnect_patch.assert_called_once_with("uiz")
        gconnect_patch().worksheet.assert_called_with("actions_brut")
        gconnect_patch().worksheet().append_row.assert_called_once()
        appened = gconnect_patch().worksheet().append_row.call_args.args[0]
        self.assertEqual(["Joueur1", "role7", "camp8"], appened[1:4])
        self.assertIn("ouiZ", appened[4])
        self.assertIn("dZ1", appened[4])
        self.assertIn("nonZ", appened[4])
        self.assertIn("dZ2", appened[4])
        self.assertNotIn("lalaZ", appened[4])



class TestVoterAgir(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.voter_agir commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = voter_agir.VoterAgir(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()
        

    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.voter_agir.export_vote")    # tested before
    async def test_vote(self, export_patch):
        """Unit tests for !vote command."""
        # async def vote(self, ctx, *, cible=None)
        vote = self.cog.vote
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            votant_village=False, vote_condamne_="oh")
        joueur.add()

        # votant_village False
        ctx = mock_discord.get_ctx(vote, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas autorisé", ctx.send.call_args.args[0])
        export_patch.assert_not_called()

        # no vote
        joueur.votant_village = True
        joueur.vote_condamne_ = None
        joueur.update()
        ctx = mock_discord.get_ctx(vote, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas de vote", ctx.send.call_args.args[0].lower())
        export_patch.assert_not_called()

        # not haroted
        joueur.vote_condamne_ = "oh"
        joueur.update()
        bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2").add()
        ctx = mock_discord.get_ctx(vote, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(("yes_no", False)):
            # abort (answer "no" at non-haroted warning)
            await ctx.invoke()
        calls = ctx.send.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertIn("n'a pas (encore) subi", calls[0].args[0])
        self.assertIn("mission aborted", calls[1].args[0])
        export_patch.assert_not_called()

        # closed during haro check
        def close_vote(_):
            joueur.vote_condamne_ = None
            joueur.update()
            return True
        ctx = mock_discord.get_ctx(vote, cible="Joueur2", _caller_id=1)
        with mock.patch("lgrez.blocs.tools.yes_no", side_effect=close_vote):
            # close vote and answer "yes" at non-haroted warning
            await ctx.invoke()
        calls = ctx.send.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertIn("a fermé entre temps", calls[1].args[0])
        export_patch.assert_not_called()

        # ok
        joueur.vote_condamne_ = "oh"        # reopen vote
        joueur.update()
        bdd.CandidHaro(_joueur_id=2, type=bdd.CandidHaroType.haro).add()
        ctx = mock_discord.get_ctx(vote, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_condamne_, "Joueur2")
        export_patch.assert_called_once_with("cond", joueur)
        export_patch.reset_mock()

        # ok, cible not specified
        joueur.vote_condamne_ = "oh"        # reset vote
        joueur.update()
        ctx = mock_discord.get_ctx(vote, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("Joueur2"))):
            await ctx.invoke()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_condamne_, "Joueur2")
        export_patch.assert_called_once_with("cond", joueur)
        export_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.voter_agir.export_vote")    # tested before
    async def test_votemaire(self, export_patch):
        """Unit tests for !votemaire command."""
        # async def votemaire(self, ctx, *, cible=None)
        votemaire = self.cog.votemaire
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            votant_village=False, vote_maire_="oh")
        joueur.add()
        export_patch.assert_not_called()

        # votant_village False
        ctx = mock_discord.get_ctx(votemaire, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas autorisé", ctx.send.call_args.args[0])
        export_patch.assert_not_called()

        # no votemaire
        joueur.votant_village = True
        joueur.vote_maire_ = None
        joueur.update()
        ctx = mock_discord.get_ctx(votemaire, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas de vote", ctx.send.call_args.args[0].lower())
        export_patch.assert_not_called()

        # not candid
        joueur.vote_maire_ = "oh"
        joueur.update()
        bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2").add()
        ctx = mock_discord.get_ctx(votemaire, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(("yes_no", False)):
            # abort (answer "no" at non-haroted warning)
            await ctx.invoke()
        calls = ctx.send.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertIn("ne s'est pas (encore) présenté", calls[0].args[0])
        self.assertIn("mission aborted", calls[1].args[0])
        export_patch.assert_not_called()

        # closed during haro check
        def close_vote(_):
            joueur.vote_maire_ = None
            joueur.update()
            return True
        ctx = mock_discord.get_ctx(votemaire, cible="Joueur2", _caller_id=1)
        with mock_discord.interact(("yes_no", close_vote)):
            # close votemaire and answer "yes" at non-haroted warning
            await ctx.invoke()
        calls = ctx.send.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertIn("a fermé entre temps", calls[1].args[0])
        export_patch.assert_not_called()

        # ok
        joueur.vote_maire_ = "oh"        # reopen votemaire
        joueur.update()
        bdd.CandidHaro(_joueur_id=2, type=bdd.CandidHaroType.candidature).add()
        ctx = mock_discord.get_ctx(votemaire, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_maire_, "Joueur2")
        export_patch.assert_called_once_with("maire", joueur)
        export_patch.reset_mock()

        # ok, cible not specified
        joueur.vote_maire_ = "oh"        # reset votemaire
        joueur.update()
        ctx = mock_discord.get_ctx(votemaire, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("Joueur2"))):
            await ctx.invoke()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_maire_, "Joueur2")
        export_patch.assert_called_once_with("maire", joueur)
        export_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.voter_agir.export_vote")    # tested before
    async def test_voteloups(self, export_patch):
        """Unit tests for !voteloups command."""
        # async def voteloups(self, ctx, *, cible=None)
        voteloups = self.cog.voteloups
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            votant_loups=False, vote_loups_="oh")
        joueur.add()

        # votant_loups False
        ctx = mock_discord.get_ctx(voteloups, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas autorisé", ctx.send.call_args.args[0])
        export_patch.assert_not_called()

        # no voteloups
        joueur.votant_loups = True
        joueur.vote_loups_ = None
        joueur.update()
        ctx = mock_discord.get_ctx(voteloups, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas de vote", ctx.send.call_args.args[0].lower())
        export_patch.assert_not_called()

        # closed during cible choice
        joueur.vote_loups_ = "oh"
        joueur.update()
        cible = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2")
        cible.add()

        ctx = mock_discord.get_ctx(voteloups, _caller_id=1)

        def close_vote(*args, **kwargs):
            joueur.vote_loups_ = None
            joueur.update()
            return ctx.new_message("Joueur2")

        with mock_discord.interact(("wait_for_message_here", close_vote)):
            # close voteloups and return cible
            await ctx.invoke()
        self.assertEqual(ctx.send.call_count, 2)
        self.assertIn("a fermé entre temps", ctx.send.call_args.args[0])
        export_patch.assert_not_called()

        # ok
        joueur.vote_loups_ = "oh"        # reopen voteloups
        joueur.update()
        ctx = mock_discord.get_ctx(voteloups, cible="Joueur2", _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_loups_, "Joueur2")
        export_patch.assert_called_once_with("loups", joueur)
        export_patch.reset_mock()

        # ok, cible not specified
        joueur.vote_loups_ = "oh"        # reset voteloups
        joueur.update()
        ctx = mock_discord.get_ctx(voteloups, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("zzz")),
                ("wait_for_message_here", ctx.new_message("Joueur2"))):
            await ctx.invoke()
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertIn("bien pris en compte", ctx.send.call_args.args[0])
        self.assertEqual(joueur.vote_loups_, "Joueur2")
        export_patch.assert_called_once_with("loups", joueur)
        export_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.gestion_actions.close_action")
    @mock.patch("lgrez.features.voter_agir.export_vote")    # tested before
    async def test_action(self, export_patch, ca_patch):
        """Unit tests for !action command."""
        # async def action(self, ctx, *, decision=None)
        action = self.cog.action
        mock_bdd.add_campsroles()
        joueur = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                            role_actif=False)
        joueur.add()

        # role_actif False
        ctx = mock_discord.get_ctx(action, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("ne peux pas utiliser", ctx.send.call_args.args[0])
        export_patch.assert_not_called()
        ca_patch.assert_not_called()

        # no actions
        joueur.role_actif = True
        joueur.update()
        ctx = mock_discord.get_ctx(action, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Aucune action en cours", ctx.send.call_args.args[0])
        export_patch.assert_not_called()
        ca_patch.assert_not_called()

        # 1 closed action
        bdd.BaseAction(slug="ouiZ").add()
        action1 = bdd.Action(joueur=joueur, _base_slug="ouiZ")
        action1.add()
        ctx = mock_discord.get_ctx(action, _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("Aucune action en cours", ctx.send.call_args.args[0])
        export_patch.assert_not_called()
        ca_patch.assert_not_called()

        # 1 open action and decision_
        action1.decision_ = "oh"
        action1.update()
        ctx = mock_discord.get_ctx(action, decision="boo", _caller_id=1)
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("boo", ctx.send.call_args.args[0])
        self.assertIn("bien prise en compte", ctx.send.call_args.args[0])
        self.assertIn("ouiZ", ctx.send.call_args.args[0])
        self.assertEqual(action1.decision_, "boo")
        export_patch.assert_called_once_with("action", joueur)
        export_patch.reset_mock()
        ca_patch.assert_not_called()

        # 1 open action and no decision_
        action1.decision_ = "oh"
        action1.update()
        ctx = mock_discord.get_ctx(action, _caller_id=1)
        with mock_discord.interact(
                ("wait_for_message_here", ctx.new_message("boo"))):
            await ctx.invoke()
        ctx.send.assert_called()
        self.assertIn("boo", ctx.send.call_args.args[0])
        self.assertIn("bien prise en compte", ctx.send.call_args.args[0])
        self.assertIn("ouiZ", ctx.send.call_args.args[0])
        self.assertEqual(action1.decision_, "boo")
        export_patch.assert_called_once_with("action", joueur)
        export_patch.reset_mock()
        ca_patch.assert_not_called()

        # 1 open INSTANT action and abort
        action1.decision_ = "oh"
        action1.base.instant = True
        action1.update()
        ctx = mock_discord.get_ctx(action, decision="boo", _caller_id=1)
        with mock_discord.interact(("yes_no", False)):
            # abort (answer "no" at instant warning)
            await ctx.invoke()
        ctx.send.assert_called()
        calls = ctx.send.call_args_list
        self.assertIn("conséquence instantanée", calls[0].args[0])
        self.assertIn("Ça part ?", calls[0].args[0])
        self.assertIn("aborted", calls[1].args[0])
        self.assertEqual(action1.decision_, "oh")
        export_patch.assert_not_called()
        ca_patch.assert_not_called()

        # 1 open INSTANT action and proceed
        ctx = mock_discord.get_ctx(action, decision="boo", _caller_id=1)
        with mock_discord.interact(("yes_no", True)):
            # proceed (answer "yes" at instant warning)
            await ctx.invoke()
        ctx.assert_sent("Attention", [str(config.Role.mj.mention),
                                      "conséquence instantanée"])
        self.assertEqual(action1.decision_, "boo")
        export_patch.assert_called_once_with("action", joueur)
        export_patch.reset_mock()
        ca_patch.assert_called_once_with(action1)
        ca_patch.reset_mock()

        # 1 open action and closed during decision choice
        action1.decision_ = "oh"
        action1.base.instant = False
        action1.update()

        def close_action(*args, **kwargs):
            action1.decision_ = None
            action1.update()
            return ctx.new_message("boo")
        ctx = mock_discord.get_ctx(action, _caller_id=1)
        with mock_discord.interact(("wait_for_message_here", close_action)):
            # close action and return decision
            await ctx.invoke()
        ctx.send.assert_called()
        self.assertIn("a fermé entre temps", ctx.send.call_args.args[0])
        export_patch.assert_not_called()
        ca_patch.assert_not_called()

        # 2 open actions and decision_: ask
        action1.decision_ = "oh"
        action1.update()
        bdd.BaseAction(slug="nonZ").add()
        action2 = bdd.Action(joueur=joueur, _base_slug="nonZ", decision_="uh")
        action2.add()
        ctx = mock_discord.get_ctx(action, decision="ih", _caller_id=1)

        with mock_discord.interact(
                ("choice", 2),
                ("wait_for_message_here", ctx.new_message("boo"))):
            await ctx.invoke()
        ctx.send.assert_called()
        calls = ctx.send.call_args_list
        self.assertIn("ouiZ", calls[0].args[0])
        self.assertIn("nonZ", calls[0].args[0])
        self.assertIn("Pour laquelle", calls[0].args[0])
        self.assertIn("nonZ", calls[1].args[0])
        self.assertIn("veux-tu faire", calls[1].args[0])
        self.assertIn("boo", calls[2].args[0])
        self.assertIn("bien prise en compte", calls[2].args[0])
        self.assertIn("nonZ", calls[2].args[0])
        self.assertEqual(action1.decision_, "oh")
        self.assertEqual(action2.decision_, "boo")
        export_patch.assert_called_once_with("action", joueur)
        ca_patch.assert_not_called()



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

import asyncio
import datetime
import time

import sqlalchemy
from sqlalchemy.ext.hybrid import hybrid_property

from lgrez import config
from lgrez.bdd import base, ActionTrigger
from lgrez.bdd.base import (autodoc_Column, autodoc_ManyToOne,
                            autodoc_OneToMany, autodoc_DynamicOneToMany)
from lgrez.bdd.enums import UtilEtat, CibleType, Vote


# Tables de données

class Action(base.TableBase):
    """Table de données des actions attribuées (liées à un joueur).

    Les instances doivent être enregistrées via
    :func:`.gestion_actions.add_action` et supprimées via
    :func:`.gestion_actions.delete_action`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique de l'action, sans signification")

    _joueur_id = sqlalchemy.Column(sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False)
    joueur = autodoc_ManyToOne("Joueur", back_populates="actions",
        doc="Joueur concerné")

    _base_slug = sqlalchemy.Column(sqlalchemy.ForeignKey("baseactions.slug"))
    base = autodoc_ManyToOne("BaseAction", back_populates="actions",
        nullable=True,
        doc="Action de base (``None`` si action de vote)")

    vote = autodoc_Column(sqlalchemy.Enum(Vote),
        doc="Si action de vote, vote concerné")

    active = autodoc_Column(sqlalchemy.Boolean(), nullable=False, default=True,
        doc="Si l'action est actuellement utilisable (False = archives)")

    cooldown = autodoc_Column(sqlalchemy.Integer(), nullable=False, default=0,
        doc="Nombre d'ouvertures avant disponiblité de l'action")
    charges = autodoc_Column(sqlalchemy.Integer(),
        doc="Nombre de charges restantes (``None`` si illimité)")

    # One-to-manys
    taches = autodoc_OneToMany("Tache", back_populates="action",
        doc="Tâches liées à cette action")

    utilisations = autodoc_DynamicOneToMany("Utilisation",
        back_populates="action",
        doc="Utilisations de cette action")

    def __init__(self, *args, **kwargs):
        """Initialize self."""
        n_args = (("base" in kwargs) + ("_base_slug" in kwargs)
                  + ("vote" in kwargs))
        if not n_args:
            raise ValueError("bdd.Action: 'base'/'_base_slug' or 'vote' "
                             "keyword-only argument must be specified")
        elif n_args > 1:
            raise ValueError("bdd.Action: 'base'/'_base_slug' and 'vote' "
                             "keyword-only argument cannot both be specified")
        super().__init__(*args, **kwargs)

    def __repr__(self):
        """Return repr(self)."""
        return f"<Action #{self.id} ({self.base or self.vote}/{self.joueur})>"

    @property
    def utilisation_ouverte(self):
        """:class:`~bdd.Utilisation` | ``None``: Utilisation de l'action
        actuellement ouverte.

        Vaut ``None`` si aucune action n'a actuellement l'état
        :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        filtre = Utilisation.etat.in_({UtilEtat.ouverte, UtilEtat.remplie})
        try:
            return self.utilisations.filter(filtre).one_or_none()
        except sqlalchemy.orm.exc.MultipleResultsFound:
            raise ValueError(
                f"Plusieurs utilisations ouvertes pour `{self}` !"
            )

    @property
    def derniere_utilisation(self):
        """:class:`~bdd.Utilisation` | ``None``:: Dernière utilisation de
        cette action (temporellement).

        Considère l'utilisation ouverte le cas échéant, sinon la
        dernière utilisation par timestamp de fermeture descendant
        (quelque soit son état, y comprs :attr:`~.bdd.UtilEtat.contree`).

        Vaut ``None`` si l'action n'a jamais été utilisée.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        return (
            self.utilisation_ouverte
            or self.utilisations.order_by(Utilisation.ts_close.desc()).first()
        )

    @property
    def decision(self):
        """str: Description de la décision de la dernière utilisation.

        Considère l'utilisation ouverte le cas échéant, sinon la
        dernière utilisation par timestamp de fermeture descendant.

        Vaut :attr:`.Utilisation.decision`, ou ``"<N/A>"`` si il n'y a
        aucune utilisation de cette action.

        Raises:
            RuntimeError: plus d'une action a actuellement l'état
            :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.remplie`.
        """
        util = self.derniere_utilisation
        if util:
            return util.decision
        else:
            return "<N/A>"

    @hybrid_property
    def is_open(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        L'action est ouverte (l'utilisateur peut interagir) ?

        *I.e.* l'action a au moins une utilisation
        :attr:`~.bdd.UtilEtat.ouverte` ou :attr:`~.bdd.UtilEtat.remplie`.

        Propriété hybride (:class:`sqlalchemy.ext.hybrid.hybrid_property`) :

            - Sur l'instance, renvoie directement la valeur booléenne ;
            - Sur la classe, renvoie la clause permettant de déterminer
              si l'action est en attente.

        Examples::

            action.is_open          # bool
            Joueur.query.filter(Joueur.actions.any(Action.is_open)).all()
        """
        return bool(self.utilisations.filter(Utilisation.is_open).all())

    @is_open.expression
    def is_open(cls):
        return cls.utilisations.any(Utilisation.is_open)

    @hybrid_property
    def is_waiting(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        L'action est ouverte et aucune décision n'a été prise ?

        *I.e.* la clause a au moins une utilisation
        :attr:`~.bdd.UtilEtat.ouverte`.

        Propriété hybride (voir :attr:`.is_open` pour plus d'infos)
        """
        return bool(self.utilisations.filter(Utilisation.is_waiting).all())

    @is_waiting.expression
    def is_waiting(cls):
        return cls.utilisations.any(Utilisation.is_waiting)


class Utilisation(base.TableBase):
    """Table de données des utilisations des actions.

    Les instances sont enregistrées via :meth:`\!open
    <.open_close.OpenClose.OpenClose.open.callback>` ;
    elles n'ont pas vocation à être supprimées.
    """
    id = autodoc_Column(sqlalchemy.BigInteger(), primary_key=True,
        doc="Identifiant unique de l'utilisation, sans signification")

    _action_id = sqlalchemy.Column(sqlalchemy.ForeignKey("actions.id"),
        nullable=False)
    action = autodoc_ManyToOne("Action", back_populates="utilisations",
        doc="Action utilisée")

    etat = autodoc_Column(sqlalchemy.Enum(UtilEtat), nullable=False,
        default=UtilEtat.ouverte,
        doc="État de l'utilisation")

    ts_open = autodoc_Column(sqlalchemy.DateTime(),
        doc="Timestamp d'ouverture de l'utilisation")
    ts_close = autodoc_Column(sqlalchemy.DateTime(),
        doc="Timestamp de fermeture de l'utilisation")
    ts_decision = autodoc_Column(sqlalchemy.DateTime(),
        doc="Timestamp du dernier remplissage de l'utilisation")

    # One-to-manys
    ciblages = autodoc_OneToMany("Ciblage", back_populates="utilisation",
        doc="Cibles désignées dans cette utilisation")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Utilisation #{self.id} ({self.action}/{self.etat})>"

    def open(self):
        """Ouvre cette utilisation.

        Modifie son :attr:`etat`, définit :attr:`ts_open` au temps
        actuel, et update.
        """
        self.etat = UtilEtat.ouverte
        self.ts_open = datetime.datetime.now()
        self.update()

    def close(self):
        """Clôture cette utilisation.

        Modifie son :attr:`etat`, définit :attr:`ts_close` au temps
        actuel, et update.
        """
        if self.etat == UtilEtat.remplie:
            self.etat = UtilEtat.validee
        else:
            self.etat = UtilEtat.ignoree
        self.ts_close = datetime.datetime.now()
        self.update()

    def ciblage(self, slug):
        """Renvoie le ciblage de base de slug voulu.

        Args:
            slug (str): Doit correspondre à un des slugs des bases
                des :attr:`ciblages` de l'utilisation.

        Returns:
            :class:`.bdd.Ciblage`

        Raises:
            ValueError: slug non trouvé dans les :attr:`ciblages`
        """
        try:
            return next(cib for cib in self.ciblages if cib.base.slug == slug)
        except StopIteration:
            raise ValueError(
                f"{self} : pas de ciblage de slug '{slug}'"
            ) from None

    @property
    def cible(self):
        """:class:`~bdd.Joueur` | ``None``: Joueur ciblé par l'utilisation,
        si applicable.

        Cet attribut n'est accessible que si l'utilisation est d'un vote
        ou d'une action définissant un et une seul ciblage de type
        :attr:`~bdd.CibleType.joueur`, :attr:`~bdd.CibleType.vivant`
        ou :attr:`~bdd.CibleType.mort`.

        Vaut ``None`` si l'utilisation a l'état
        :attr:`~bdd.UtilEtat.ouverte` ou :attr:`~bdd.UtilEtat.ignoree`.

        Raises:
            ValueError: l'action ne remplit pas les critères évoqués
            ci-dessus
        """
        if self.action.vote:
            # vote : un BaseCiblage implicite de type CibleType.vivants
            return self.ciblages[0].joueur if self.ciblages else None
        else:
            base_ciblages = self.action.base.base_ciblages
            bc_joueurs = [bc for bc in base_ciblages
                          if bc.type in [CibleType.joueur, CibleType.vivant,
                                         CibleType.mort]]
            if len(bc_joueurs) != 1:
                raise ValueError (f"L'utilisation {self} n'a pas une et "
                                  "une seule cible de type joueur")

            base_ciblage = bc_joueurs[0]
            try:
                ciblage = next(cib for cib in self.ciblages
                               if cib.base == base_ciblage)
            except StopIteration:
                return None         # Pas de ciblage fait

            return ciblage.joueur

    @property
    def decision(self):
        """str: Description de la décision de cette utilisation.

        Complète le template de :.bdd.BaseAction.decision_format` avec
        les valeurs des ciblages de l'utilisation.

        Vaut ``"Ne rien faire"`` si l'utilisation n'a pas de ciblages,
        et :attr:`.cible` dans le cas d'un vote.
        """
        if not self.action.base:
            return str(self.cible)

        if not self.ciblages:
            return "Ne rien faire"

        template = self.action.base.decision_format
        data = {ciblage.base.slug: ciblage.valeur_descr
                for ciblage in self.ciblages}
        try:
            return template.format(**data)
        except KeyError:
            return template

    @hybrid_property
    def is_open(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        L'utilisation est ouverte (l'utilisateur peut interagir) ?

        Raccourci pour
        ``utilisation.etat in {UtilEtat.ouverte, UtilEtat.remplie}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return (self.etat in {UtilEtat.ouverte, UtilEtat.remplie})

    @is_open.expression
    def is_open(cls):
        return cls.etat.in_({UtilEtat.ouverte, UtilEtat.remplie})

    @hybrid_property
    def is_waiting(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        L'utilisation est ouverte et aucune décision n'a été prise ?

        Raccourci pour ``utilisation.etat == UtilEtat.ouverte``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return (self.etat == UtilEtat.ouverte)

    @hybrid_property
    def is_filled(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        L'utilisation est remplie (l'utilisateur a interagi avec) ?

        Raccourci pour
        ``utilisation.etat in {UtilEtat.remplie, UtilEtat.validee,
        UtilEtat.contree}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return (self.etat in {UtilEtat.remplie, UtilEtat.validee,
                              UtilEtat.contree})

    @is_filled.expression
    def is_filled(cls):
        return cls.etat.in_({UtilEtat.remplie, UtilEtat.validee,
                             UtilEtat.contree})


class Ciblage(base.TableBase):
    """Table de données des cibles désignées dans les utilisations d'actions.

    Les instances sont enregistrées via :meth:`\!action
    <.voter_agir.VoterAgir.VoterAgir.action.callback>` ;
    elles n'ont pas vocation à être supprimées.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique du ciblage, sans signification")

    _base_id = sqlalchemy.Column(sqlalchemy.ForeignKey("baseciblages._id"))
    base = autodoc_ManyToOne("BaseCiblage", back_populates="ciblages",
        nullable=True,
        doc="Modèle de ciblage (lié au modèle d'action). Vaut ``None`` pour "
            "un ciblage de vote")

    _utilisation_id = sqlalchemy.Column(sqlalchemy.ForeignKey(
        "utilisations.id"), nullable=False)
    utilisation = autodoc_ManyToOne("Utilisation", back_populates="ciblages",
        doc="Utilisation où ce ciblage a été fait")

    _joueur_id = sqlalchemy.Column(sqlalchemy.ForeignKey(
        "joueurs.discord_id"), nullable=True)
    joueur = autodoc_ManyToOne("Joueur", back_populates="ciblages",
        nullable=True, doc="Joueur désigné, si ``base.type`` vaut "
        ":attr:`~.bdd.CibleType.joueur`, :attr:`~.bdd.CibleType.vivant` "
        "ou :attr:`~.bdd.CibleType.mort`")

    _role_slug = sqlalchemy.Column(sqlalchemy.ForeignKey(
        "roles.slug"), nullable=True)
    role = autodoc_ManyToOne("Role", back_populates="ciblages",
        nullable=True, doc="Rôle désigné, si ``base.type`` vaut "
        ":attr:`~.bdd.CibleType.role`")

    _camp_slug = sqlalchemy.Column(sqlalchemy.ForeignKey(
        "camps.slug"), nullable=True)
    camp = autodoc_ManyToOne("Camp", back_populates="ciblages",
        nullable=True, doc="Camp désigné, si ``base.type`` vaut "
        ":attr:`~.bdd.CibleType.camp`")

    booleen = autodoc_Column(sqlalchemy.Boolean(),
        doc="Valeur, si ``base.type`` vaut :attr:`~.bdd.CibleType.booleen`")
    texte = autodoc_Column(sqlalchemy.String(1000),
        doc="Valeur, si ``base.type`` vaut :attr:`~.bdd.CibleType.texte`")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Ciblage #{self.id} ({self.base}/{self.utilisation})>"

    @property
    def _val_attr(self):
        """Nom de l'attribut stockant la valeur du ciblage"""
        if (not self.base       # vote
            or self.base.type in {CibleType.joueur, CibleType.vivant,
                                  CibleType.mort}):
            return "joueur"
        elif self.base.type == CibleType.role:
            return "role"
        elif self.base.type == CibleType.camp:
            return "camp"
        elif self.base.type == CibleType.booleen:
            return "booleen"
        elif self.base.type == CibleType.texte:
            return "texte"
        else:
            raise ValueError(f"Ciblage de type inconnu : {self.base.type}")

    @property
    def valeur(self):
        """:class:`~bdd.Joueur` | :class:`~bdd.Role`| :class:`~bdd.Camp`
        | :class:`bool` | :class:`str`: Valeur du ciblage, selon son type.

        Propriété en lecture et écriture.

        Raises:
            ValueError: ciblage de type inconnu
        """
        return getattr(self, self._val_attr)

    @valeur.setter
    def valeur(self, value):
        setattr(self, self._val_attr, value)

    @property
    def valeur_descr(self):
        """str: Description de la valeur du ciblage.

        Si :attr:`valeur` vaut ``None``, renvoie ``<N/A>``

        Raises:
            ValueError: ciblage de type inconnu
        """
        if self.valeur is None:
            return "<N/A>"

        if (not self.base       # vote
            or self.base.type in {CibleType.joueur, CibleType.vivant,
                                  CibleType.mort}):
            return self.joueur.nom
        elif self.base.type == CibleType.role:
            return self.role.nom_complet
        elif self.base.type == CibleType.camp:
            return self.camp.nom
        elif self.base.type == CibleType.booleen:
            return "Oui" if self.booleen else "Non"
        else:
            return self.texte


class Tache(base.TableBase):
    """Table de données des tâches planifiées du bot.

    Les instances doivent être enregistrées via :meth:`.add`
    et supprimées via :func:`.delete`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique de la tâche, sans signification")
    timestamp = autodoc_Column(sqlalchemy.DateTime(), nullable=False,
        doc="Moment où exécuter la tâche")
    commande = autodoc_Column(sqlalchemy.String(2000), nullable=False,
        doc="Texte à envoyer via le webhook (généralement une commande)")

    _action_id = sqlalchemy.Column(sqlalchemy.ForeignKey("actions.id"),
        nullable=True)
    action = autodoc_ManyToOne("Action", back_populates="taches",
        nullable=True,
        doc="Si la tâche est liée à une action, action concernée")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Tache #{self.id} ({self.commande})>"

    @property
    def handler(self):
        """asyncio.TimerHandle: Représentation dans le bot de la tâche.

        Proxy pour :attr:`config.bot.tasks[self.id] <.LGBot.tasks>`,
        en lecture, écriture et suppression (``del``).

        Raises:
            RuntimeError: tâche non enregistrée dans le bot.
        """
        try:
            return config.bot.tasks[self.id]
        except KeyError:
            raise RuntimeError(f"Tâche {self} non enregistrée dans le bot !")

    @handler.setter
    def handler(self, value):
        if self.id is None:
            raise RuntimeError("Tache.handler: Tache.id non défini (commit ?)")
        config.bot.tasks[self.id] = value

    @handler.deleter
    def handler(self):
        try:
            del config.bot.tasks[self.id]
        except KeyError:
            pass

    async def send_webhook(self, tries=0):
        """Exécute la tâche (coroutine programmée par :meth:`execute`).

        Envoie un webhook (:obj:`.config.webhook`) de contenu
        :attr:`commande`.

        Si une exception quelconque est levée par l'envoi du webhook,
        re-programme l'exécution de la tâche (:meth:`execute`) 2 secondes
        après ; ce jusqu'à 5 fois, après quoi un message d'alerte est
        envoyé dans :attr:`.config.Channel.logs`.

        Si aucune exception n'est levée (succès), supprime la tâche.

        Args:
            tries (int): Numéro de l'essai d'envoi actuellement en cours
        """
        try:
            await config.webhook.send(self.commande)
        except Exception as exc:
            if tries < 5:
                # On réessaie
                config.loop.call_later(2, self.execute, tries + 1)
            else:
                await config.Channel.logs.send(
                    f"{config.Role.mj.mention} ALERT: impossible  "
                    f"d'envoyer un webhook (5 essais, erreur : "
                    f"```{type(exc).__name__}: {exc})```\n"
                    f"Commande non envoyée : `{self.commande}`"
                )
        else:
            self.delete()

    def execute(self, tries=0):
        """Exécute la tâche planifiée (méthode appellée par la loop).

        Programme :meth:`send_webhook` pour exécution immédiate.

        Args:
            tries (int): Numéro de l'essai d'envoi actuellement en cours,
                passé à :meth:`send_webhook`.
        """
        asyncio.create_task(self.send_webhook(tries=tries))
        # programme la coroutine pour exécution immédiate

    def register(self):
        """Programme l'exécution de la tâche dans la loop du bot."""
        now = datetime.datetime.now()
        delay = (self.timestamp - now).total_seconds()
        TH = config.loop.call_later(delay, self.execute)
        # Programme la tâche (appellera tache.execute() à timestamp)
        self.handler = TH               # TimerHandle, pour pouvoir cancel

    def cancel(self):
        """Annule et nettoie la tâche planifiée (sans la supprimer en base).

        Si la tâche a déjà été exécutée, ne fait que nettoyer le handler.
        """
        try:
            self.handler.cancel()       # Annule la task (objet TimerHandle)
            # (pas d'effet si la tâche a déjà été exécutée)
        except RuntimeError:            # Tache non enregistrée
            pass
        else:
            del self.handler

    def add(self, *other):
        """Enregistre la tâche sur le bot et en base.

        Globalement équivalent à un appel à :meth:`.register` (pour
        chaque élément le cas échéant) avant l'ajout en base habituel
        (:meth:`TableBase.add <.bdd.base.TableBase.add>`).

        Args:
            \*other: autres instances à ajouter dans le même commit,
                éventuellement.
        """
        super().add(*other)             # Enregistre tout en base

        self.register()                 # Enregistre sur le bot
        for item in other:              # Les autres aussi
            item.register()


    def delete(self, *other):
        """Annule la tâche planifiée et la supprime en base.

        Globalement équivalent à un appel à :meth:`.cancel` (pour
        chaque élément le cas échéant) avant la suppression en base
        habituelle (:meth:`TableBase.add <.bdd.base.TableBase.add>`).

        Args:
            \*other: autres instances à supprimer dans le même commit,
                éventuellement.
        """
        self.cancel()                   # Annule la tâche
        for item in other:              # Les autres aussi
            item.cancel()

        super().delete(*other)          # Supprime tout en base



import unittest
from unittest import mock

from lgrez.blocs import webhook



class TestWebhook(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.webhook module."""

    @mock.patch("discord_webhook.DiscordWebhook")
    def test_send(self, wh_patch):
        """Unit tests for webhook.send function."""
        # def send(message, url)
        send = webhook.send
        # unique case
        message = mock.Mock()
        url = mock.Mock()
        rep = send(message, url)
        wh_patch.assert_called_once_with(content=message, url=url)
        wh_patch.return_value.execute.assert_called_once()
        self.assertEqual(rep, wh_patch.return_value.execute.return_value)



"""lg-rez / bdd / Modèle de données

Déclaration de toutes les tables et leurs colonnes

"""

import datetime

import sqlalchemy
from sqlalchemy.ext.hybrid import hybrid_property

from lgrez import config
from lgrez.bdd import base
from lgrez.bdd.base import (autodoc_Column, autodoc_ManyToOne,
                            autodoc_OneToMany, autodoc_DynamicOneToMany)
from lgrez.bdd.enums import Statut, CandidHaroType, Vote


# Tables de données

class Joueur(base.TableBase):
    """Table de données des joueurs inscrits.

    Les instances de cette classe correspondent aux lignes du Tableau de
    bord ; elles sont crées par l'inscription (:func:`.inscription.main`)
    et synchronisées par :meth:`\!sync <.sync.Sync.Sync.sync.callback>`.
    """
    discord_id = autodoc_Column(sqlalchemy.BigInteger(), primary_key=True,
        autoincrement=False, doc="ID Discord du joueur")
    chan_id_ = autodoc_Column(sqlalchemy.BigInteger(), nullable=False,
        doc="ID du chan privé Discord du joueur. *(le ``_`` final indique "
            "que ce champ n'est pas synchnisé avec le Tableau de bord)*")

    nom = autodoc_Column(sqlalchemy.String(32), nullable=False,
        doc="Nom du joueur (demandé à l'inscription)")
    chambre = autodoc_Column(sqlalchemy.String(200),
        doc="Emplacement du joueur (demandé à l'inscription)")
    statut = autodoc_Column(sqlalchemy.Enum(Statut), nullable=False,
        default=Statut.vivant, doc="Statut RP")

    _role_slug = sqlalchemy.Column(sqlalchemy.ForeignKey("roles.slug"),
        nullable=False, default=lambda: config.default_role_slug)
    role = autodoc_ManyToOne("Role", back_populates="joueurs",
        doc="Rôle du joueur (défaut :meth:`.bdd.Role.default`)")

    _camp_slug = sqlalchemy.Column(sqlalchemy.ForeignKey("camps.slug"),
        nullable=False, default=lambda: config.default_camp_slug)
    camp = autodoc_ManyToOne("Camp", back_populates="joueurs",
        doc="Camp du joueur (défaut :meth:`.bdd.Camp.default`)")

    votant_village = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=True, doc="Le joueur participe aux votes du village ?")
    votant_loups = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=False, doc="Le joueur participe au vote des loups ?")
    role_actif = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=True, doc="Le joueur peut agir ? (pas chatgarouté...)")

    # One-to-manys
    actions = autodoc_OneToMany("Action", back_populates="joueur",
        doc="Actions pour ce joueur")
    candidharos = autodoc_OneToMany("CandidHaro", back_populates="joueur",
        doc="Candidatures et haros de/contre ce joueur")
    bouderies = autodoc_OneToMany("Bouderie", back_populates="joueur",
        doc="Appartenances aux boudoirs de ce joueur")
    ciblages = autodoc_DynamicOneToMany("Ciblage", back_populates="joueur",
        doc="Ciblages prenant ce joueur pour cible")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Joueur #{self.discord_id} ({self.nom})>"

    def __str__(self):
        """Return str(self)."""
        return str(self.nom)

    @property
    def member(self):
        """discord.Member: Membre Discord correspondant à ce Joueur.

        Raises:
            ValueError: pas de membre correspondant
            ~ready_check.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_member(self.discord_id)
        if not result:
            raise ValueError(f"Joueur.member : pas de membre pour `{self}` !")

        return result

    @property
    def private_chan(self):
        """discord.TextChannel: Channel privé (conversation bot) du joueur.

        Raises:
            ValueError: pas de channel correspondant
            ~ready_check.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_channel(self.chan_id_)
        if not result:
            raise ValueError("Joueur.private_chan : "
                             f"pas de chan pour `{self}` !")

        return result

    @property
    def actions_actives(self):
        """Sequence[.bdd.Action]: Sous-ensemble de :attr:`actions` restreint
        aux actions actives.

        Élimine aussi les actions de vote (sans base).
        """
        return [ac for ac in self.actions if ac.active and not ac.vote]

    @property
    def boudoirs(self):
        """Sequence[.Boudoir]: Boudoirs où est ce jour (read-only)"""
        return [boud.boudoir for boud in self.bouderies]

    @hybrid_property
    def est_vivant(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        Le joueur est en vie ou MVtisé ?

        Raccourci pour
        ``joueur.statut in {Statut.vivant, Statut.MV}``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return (self.statut in {Statut.vivant, Statut.MV})

    @est_vivant.expression
    def est_vivant(cls):
        return cls.statut.in_({Statut.vivant, Statut.MV})

    @hybrid_property
    def est_mort(self):
        """:class:`bool` (instance)
        / :class:`sqlalchemy.sql.selectable.Exists` (classe):
        Le joueur est mort ?

        Raccourci pour ``joueur.statut == Statut.mort``

        Propriété hybride (voir :attr:`.Action.is_open` pour plus d'infos)
        """
        return (self.statut == Statut.mort)

    @classmethod
    def from_member(cls, member):
        """Récupère le Joueur (instance de BDD) lié à un membre Discord.

        Args:
            member (discord.Member): le membre concerné

        Returns:
            Joueur: Le joueur correspondant.

        Raises:
            ValueError: membre introuvable en base
            ~ready_check.NotReadyError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        joueur = cls.query.get(member.id)
        if not joueur:
            raise ValueError("Joueur.from_member : "
                             f"pas de joueur en base pour `{member}` !")

        return joueur

    def action_vote(self, vote):
        """Retourne l'"action de vote" voulue pour ce joueur.

        Args:
            vote (.bdd.Vote): vote pour lequel récupérer l'action

        Returns:
            :class:`~bdd.Action`

        Raises:
            RuntimeError: action non existante
        """
        if not isinstance(vote, Vote):
            vote = Vote[vote]
        try:
            return next(act for act in self.actions if act.vote == vote)
        except StopIteration:
            raise RuntimeError(f"{self} : pas d'action de vote {vote.name} ! "
                               "!cparti a bien été appelé ?") from None


class CandidHaro(base.TableBase):
    """Table de données des candidatures et haros en cours #PhilippeCandidHaro.

    Les instances sont enregistrées via :meth:`\!haro
    <.actions_publiques.ActionsPubliques.ActionsPubliques.haro.callback>`
    / :meth:`\!candid
    <.actions_publiques.ActionsPubliques.ActionsPubliques.candid.callback>`
    et supprimées via :meth:`\!wipe
    <.actions_publiques.ActionsPubliques.ActionsPubliques.wipe.callback>`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique du candidharo, sans signification")

    _joueur_id = sqlalchemy.Column(sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False)
    joueur = autodoc_ManyToOne("Joueur", back_populates="candidharos",
        doc="Joueur concerné (candidat ou haroté)")

    type = autodoc_Column(sqlalchemy.Enum(CandidHaroType), nullable=False,
        doc="Haro ou candidature ?")

    def __repr__(self):
        """Return repr(self)."""
        return f"<CandidHaro #{self.id} ({self.joueur}/{self.type})>"


class Boudoir(base.TableBase):
    """Table de données des boudoirs sur le serveur.

    Les instances de cette classe sont crées, modifiées et supprimées par
    :meth:`\!boudoir <.chans.GestionChans.GestionChans.boudoir.callback>`.
    """
    chan_id = autodoc_Column(sqlalchemy.BigInteger(), primary_key=True,
        autoincrement=False, doc="ID Discord du salon")

    nom = autodoc_Column(sqlalchemy.String(32), nullable=False,
        doc="Nom du boudoir (demandé à la création)")
    ts_created = autodoc_Column(sqlalchemy.DateTime(), nullable=False,
        doc="Timestamp de la création")

    # One-to-manys
    bouderies = autodoc_OneToMany("Bouderie", back_populates="boudoir",
        doc="Appartenances à ce boudoir")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Boudoir #{self.chan_id} ({self.nom})>"

    @property
    def chan(self):
        """discord.TextChannel: Salon Discord correspondant à ce boudoir.

        Raises:
            ValueError: pas de membre correspondant
            ~ready_check.NotReadyError: bot non connecté
                (:obj:`.config.guild` vaut ``None``)
        """
        result = config.guild.get_channel(self.chan_id)
        if not result:
            raise ValueError(f"Boudoir.chan : pas de chan pour {self} !")

        return result

    @property
    def joueurs(self):
        """Sequence[.Joueur]: Joueurs présents dans ce boudoir (read-only)"""
        return [boud.joueur for boud in self.bouderies]

    @property
    def gerant(self):
        """.bdd.Joueur: Membre du boudoir ayant les droits de gestion.

        Raises:
            ValueError: pas de membre avec les droits de gestion
        """
        try:
            return next(boud.joueur for boud in self.bouderies if boud.gerant)
        except StopIteration:
            raise ValueError(f"Pas de membre gérant le boudoir *{self.name}*")

    async def add_joueur(self, joueur, gerant=False):
        """Ajoute un joueur sur le boudoir.

        Crée la :class:`.Bouderie` correspondante et modifie les
        permissions du salon.

        Args:
            joueur (.Joueur): Le joueur à ajouter.
            gerant (bool): Si le joueur doit être ajouté avec les
                permissions de gérant.
        """
        now = datetime.datetime.now()
        Bouderie(boudoir=self, joueur=joueur, gerant=gerant,
                 ts_added=now, ts_promu=now if gerant else None).add()
        await self.chan.set_permissions(joueur.member, read_messages=True)

    async def remove_joueur(self, joueur):
        """Retire un joueur du boudoir.

        Supprime la :class:`.Bouderie` correspondante et modifie les
        permissions du salon.

        Args:
            joueur (.Joueur): Le joueur à ajouter.
        """
        Bouderie.query.filter_by(boudoir=self, joueur=joueur).one().delete()
        await self.chan.set_permissions(joueur.member, overwrite=None)

    @classmethod
    def from_channel(cls, channel):
        """Récupère le Boudoir (instance de BDD) lié à un salon Discord.

        Args:
            channel (discord.TextChannel): le salon concerné

        Returns:
            .Boudoir: Le salon correspondant.

        Raises:
            ValueError: boudoir introuvable en base
            ~ready_check.NotReadyError: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        boudoir = cls.query.get(channel.id)
        if not boudoir:
            raise ValueError("Boudoir.from_channel : "
                             f"pas de boudoir en base pour `{channel.mention}` !")

        return boudoir


class Bouderie(base.TableBase):
    """Table de données des appartenances aux boudoirs du serveur.

    Table d'association entre :class:`.Joueur` et :class:`.Boudoir`.

    Les instances de cette classe sont crées, modifiées et supprimées par
    :meth:`\!boudoir <.chans.GestionChans.GestionChans.boudoir.callback>`.
    """
    id = autodoc_Column(sqlalchemy.Integer(), primary_key=True,
        doc="Identifiant unique de la bouderie, sans signification")

    _joueur_id = sqlalchemy.Column(sqlalchemy.ForeignKey("joueurs.discord_id"),
        nullable=False)
    joueur = autodoc_ManyToOne("Joueur", back_populates="bouderies",
        doc="Joueur concerné")

    _boudoir_id = sqlalchemy.Column(sqlalchemy.ForeignKey("boudoirs.chan_id"),
        nullable=False)
    boudoir = autodoc_ManyToOne("Boudoir", back_populates="bouderies",
        doc="Boudoir concerné")

    gerant = autodoc_Column(sqlalchemy.Boolean(), nullable=False,
        default=False,
        doc="Si le joueur a les droits de gestion du boudoir")

    ts_added = autodoc_Column(sqlalchemy.DateTime(), nullable=False,
        doc="Timestamp de l'ajout du joueur au boudoir")
    ts_promu = autodoc_Column(sqlalchemy.DateTime(),
        doc="Timestamp de la promotion en gérant, le cas échéant")

    def __repr__(self):
        """Return repr(self)."""
        return f"<Bouderie #{self.id} ({self.joueur}/{self.boudoir})>"



"""lg-rez / blocs / Lecture des variables d'environnement

"""

import sys
import os

from dotenv import load_dotenv


# À l'import
if len(sys.argv) > 1:
    path = sys.argv[1]
else:
    path = ".env"

load_dotenv(path)


def load(VAR_NAME):
    """Lit une valeur depuis les variables d'environnement.

    Les variables sont recherchées dans les variables d'environnement ;
    à l'import de ce module, celles-ci sont chargées par
    ``dotenv.load_dotenv`` à partir du fichier passé en premier
    argument en ligne de commande (``.env`` par défaut).

    Équivaut globalement à :func:`os.getenv` suivi d'une vérification
    que la variable existe.

    Args:
        VAR_NAME (str): nom de la variable à charger (``LGREZ_...``)

    Returns:
        :class:`str`

    Raises:
        RuntimeError: la variable d'environnement n'est pas définie
    """
    var = os.getenv(VAR_NAME)
    if var is None:
        raise RuntimeError(f"Variable d'environnement {VAR_NAME} manquante")
    return var



import asyncio
import functools

from lgrez import bdd, config
from test import mock_env


class _PatchDB():
    def __enter__(self):
        self.old_session = config.session
        self.old_engine = config.engine
        env_patch = mock_env.patch_env(LGREZ_DATABASE_URI="sqlite://")
        self.env_patcher = env_patch.__enter__()
        bdd.connect()
        return self

    def __exit__(self, etype=None, eval=None, tb=None):
        config.session.close()
        config.engine.dispose()
        self.env_patcher.__exit__(etype, eval, tb)
        config.session = self.old_session
        config.engine = self.old_engine

    def __call__(self, func):
        if asyncio.iscoroutinefunction(func):
            @functools.wraps(func)
            async def newfunc(*args, **kwargs):
                self.__enter__()
                try:
                    await func(*args, **kwargs)
                finally:
                    self.__exit__()
        else:
            @functools.wraps(func)
            def newfunc(*args, **kwargs):
                self.__enter__()
                try:
                    func(*args, **kwargs)
                finally:
                    self.__exit__()

        return newfunc


def patch_db(func=None):
    """Patch database with an in-memory empty SQLite database.

    Patch `LGREZ_DATABASE_URI` environment variable, then calls
    `bdd.connect()` to make `config.session` and `config.engine` refer
    to the newly-created database (and restores them at exit time).

    Can be used as a function decorator or context manager, as for
    `unittest.mock.patch`.
    """
    if callable(func):
        # @patch_db directly used as decorator (no parenthesis)
        return _PatchDB()(func)
    else:
        return _PatchDB()



def add_campsroles(n_camps=1, n_roles=1):
    """Ajoute `n_camps` camps et `n_roles` rôles à la BDD.

    Le premier camp/rôle ajouté est toujours celui par défaut
    (`config.default_camp/role_slug`), puis camp1, camp2...

    Retourne un tuple (liste des camps, liste des rôles).
    """
    camps = [bdd.Camp(slug=config.default_camp_slug, nom="defaut")]
    for i in range(1, n_camps):
        camps.append(bdd.Camp(slug=f"camp{i}",
                              nom=f"Camp{i}",
                              emoji=f"emoji{i}"))

    config.session.add_all(camps)

    roles = [bdd.Role(slug=config.default_role_slug, nom="defaut",
                      camp=bdd.Camp.default())]
    for i in range(1, n_roles):
        roles.append(bdd.Role(slug=f"role{i}",
                              nom=f"Role{i}",
                              camp=camps[i % n_camps]))

    config.session.add_all(roles)

    return camps, roles



import datetime
import unittest
from unittest import mock

import discord
import freezegun

from lgrez import config, bdd
from lgrez.features import taches
from test import mock_discord, mock_bdd


class TestGestionTaches(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.taches commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = taches.GestionTaches(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db      # Empty database for this method
    async def test_taches(self):
        """Unit tests for !taches command."""
        # async def taches(self, ctx)
        taches_cmd = self.cog.taches

        # no taches
        ctx = mock_discord.get_ctx(taches_cmd)
        await ctx.invoke()
        ctx.assert_sent("Aucune tâche")

        # some taches
        bdd.Tache(timestamp=datetime.datetime.now(), commande="bloup").add()
        bdd.BaseAction(slug="baaz").add()
        bdd.Joueur(discord_id=123, chan_id_=2, nom="Joueur1").add()
        bdd.Action(id=13, _base_slug="baaz", _joueur_id=123).add()
        bdd.Tache(timestamp=datetime.datetime(2021, 5, 15, 18, 0, 0),
                  commande="bliip", _action_id=13).add()
        ctx = mock_discord.get_ctx(taches_cmd)
        await ctx.invoke()
        ctx.assert_sent("Tâches en attente")
        self.assertIn("bloup", ctx.send.call_args.args[0])
        self.assertIn("bliip", ctx.send.call_args.args[0])
        self.assertIn("baaz", ctx.send.call_args.args[0])
        self.assertIn("Joueur1", ctx.send.call_args.args[0])

        # a lot of taches
        bdd.Tache.add(*(bdd.Tache(timestamp=datetime.datetime.now(),
                                  commande=f"bloup{n}") for n in range(1000)))
        ctx = mock_discord.get_ctx(taches_cmd)
        await ctx.invoke()
        sent = "\n".join(call.args[0] for call in ctx.send.call_args_list)
        for n in range(1000):
            self.assertIn(f"bloup{n}", sent)


    @mock_bdd.patch_db      # Empty database for this method
    async def test_planif(self):
        """Unit tests for !planif command."""
        # async def planif(self, ctx, quand, *, commande)
        planif = self.cog.planif
        today = datetime.date.today()
        today_12h = datetime.datetime.combine(today, datetime.time(12, 0))
        firstmarch_12h = datetime.datetime(2021, 3, 1, 12, 0)

        # bad timestamp
        ctx = mock_discord.get_ctx(planif, "bzzt", commande="oui")
        with self.assertRaises(ValueError):
            await ctx.invoke()
        self.assertEqual(bdd.Tache.query.all(), [])

        # hh:mm after now
        ctx = mock_discord.get_ctx(planif, "18:10", commande="oui")
        with freezegun.freeze_time(today_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime.combine(
            today, datetime.time(18, 10)))
        self.assertEqual(tache.commande, "oui")
        tache.delete()

        # hh:mm:ss after now
        ctx = mock_discord.get_ctx(planif, "18:10:52", commande="non")
        with freezegun.freeze_time(today_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime.combine(
            today, datetime.time(18, 10, 52)))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # hh:mm before now, abort
        ctx = mock_discord.get_ctx(planif, "10:10", commande="oui")
        with freezegun.freeze_time(today_12h):
            with mock_discord.interact(("yes_no", False)):
                await ctx.invoke()
        self.assertEqual(bdd.Tache.query.all(), [])

        # hh:mm before now, proceed
        ctx = mock_discord.get_ctx(planif, "10:10", commande="oui")
        with freezegun.freeze_time(today_12h):
            with mock_discord.interact(("yes_no", True)):
                await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime.combine(
            today, datetime.time(10, 10)))
        self.assertEqual(tache.commande, "oui")
        tache.delete()

        # jj/mm-hh:mm after now
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="non")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # j/m-h:m after now
        ctx = mock_discord.get_ctx(planif, "5/3-9:7", commande="non")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # jj/mm-hh:mm before now
        ctx = mock_discord.get_ctx(planif, "05/02-09:07", commande="non")
        with freezegun.freeze_time(firstmarch_12h):
            with mock_discord.interact(("yes_no", True)):
                await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 2, 5, 9, 7))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # jj/mm/aaaa-hh:mm after now
        ctx = mock_discord.get_ctx(planif, "05/03/2023-09:07", commande="non")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2023, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # jj/mm/aaaa-hh:mm before now
        ctx = mock_discord.get_ctx(planif, "05/03/2020-09:07", commande="non")
        with freezegun.freeze_time(firstmarch_12h):
            with mock_discord.interact(("yes_no", True)):
                await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2020, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "non")
        tache.delete()

        # jj/mm-hh:mm after now, !open <non-existing action>
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!open 77")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!open 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # jj/mm-hh:mm after now, !open <existing action>
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1").add()
        bdd.BaseAction(slug="ouiz").add()
        action = bdd.Action(id=78, _base_slug="ouiz", _joueur_id=1)
        action.add()
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!open 78")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!open 78")
        self.assertEqual(tache.action, action)
        tache.delete()

        # jj/mm-hh:mm after now, !close <non-existing action>
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!close 77")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!close 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # jj/mm-hh:mm after now, !close <existing action>
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!close 78")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!close 78")
        self.assertEqual(tache.action, action)
        tache.delete()

        # jj/mm-hh:mm after now, !remind <non-existing action>
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!remind 77")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!remind 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # jj/mm-hh:mm after now, !remind <existing action>
        ctx = mock_discord.get_ctx(planif, "05/03-09:07", commande="!remind 78")
        with freezegun.freeze_time(firstmarch_12h):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, datetime.datetime(2021, 3, 5, 9, 7))
        self.assertEqual(tache.commande, "!remind 78")
        self.assertEqual(tache.action, action)
        tache.delete()


    @mock_bdd.patch_db      # Empty database for this method
    async def test_delay(self):
        """Unit tests for !delay command."""
        # async def delay(self, ctx, duree, *, commande)
        delay = self.cog.delay
        now = datetime.datetime(2021, 3, 1, 12, 0)

        # bad duree
        bads = ["bzzt", "0h", "0m", "0s", "0h0m", "0h0s", "0m0s", "0h0m0s",
                "1h1", "1m1", "1s1", "1h1m1", "1h1s1", "1m1s1", "1m1m1s1"]
        for bad in bads:
            ctx = mock_discord.get_ctx(delay, bad, commande="oui")
            with self.assertRaises(discord.ext.commands.BadArgument):
                await ctx.invoke()
            self.assertEqual(bdd.Tache.query.all(), [])

        # diffrent ways
        oks = {
            "2h": datetime.timedelta(hours=2),
            "3m": datetime.timedelta(minutes=3),
            "5s": datetime.timedelta(seconds=5),
            "2h3m": datetime.timedelta(hours=2, minutes=3),
            "2h5s": datetime.timedelta(hours=2, seconds=5),
            "3m5s": datetime.timedelta(minutes=3, seconds=5),
            "2h3m5s": datetime.timedelta(hours=2, minutes=3, seconds=5),
        }
        for ok, delta in oks.items():
            ctx = mock_discord.get_ctx(delay, ok, commande="oui")
            with freezegun.freeze_time(now):
                await ctx.invoke()
            self.assertEqual(len(bdd.Tache.query.all()), 1)
            tache = bdd.Tache.query.one()
            self.assertEqual(tache.timestamp, now + delta)
            self.assertEqual(tache.commande, "oui")
            tache.delete()

        # ok, !open <non-existing action>
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!open 77")
        delta = oks["3m5s"]
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!open 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # ok, !open <existing action>
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1").add()
        bdd.BaseAction(slug="ouiz").add()
        action = bdd.Action(id=78, _base_slug="ouiz", _joueur_id=1)
        action.add()
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!open 78")
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!open 78")
        self.assertEqual(tache.action, action)
        tache.delete()

        # ok, !close <non-existing action>
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!close 77")
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!close 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # ok, !close <existing action>
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!close 78")
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!close 78")
        self.assertEqual(tache.action, action)
        tache.delete()

        # ok, !remind <non-existing action>
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!remind 77")
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!remind 77")
        self.assertIsNone(tache.action)
        tache.delete()

        # ok, !remind <existing action>
        ctx = mock_discord.get_ctx(delay, "3m5s", commande="!remind 78")
        with freezegun.freeze_time(now):
            await ctx.invoke()
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, now + delta)
        self.assertEqual(tache.commande, "!remind 78")
        self.assertEqual(tache.action, action)
        tache.delete()


    @mock_bdd.patch_db      # Empty database for this method
    async def test_cancel(self):
        """Unit tests for !cancel command."""
        # async def cancel(self, ctx, *ids)
        cancel = self.cog.cancel

        # no arg
        ctx = mock_discord.get_ctx(cancel)
        await ctx.invoke()
        ctx.assert_sent("Aucune tâche trouvée")

        # meaningless args
        ctx = mock_discord.get_ctx(cancel, "oh", "ezz", "O@!")
        await ctx.invoke()
        ctx.assert_sent("Aucune tâche trouvée")

        # nonexisting tasks
        ctx = mock_discord.get_ctx(cancel, "1", "2", "3")
        await ctx.invoke()
        ctx.assert_sent("Aucune tâche trouvée")

        # one existing task, abort
        bdd.Tache(id=9, timestamp=datetime.datetime.now(), commande="oz").add()
        ctx = mock_discord.get_ctx(cancel, "9")
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent("oz", "aborted")
        self.assertEqual(len(bdd.Tache.query.all()), 1)

        # one existing task, proceed
        ctx = mock_discord.get_ctx(cancel, "9")
        with mock_discord.interact(("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent("oz", "annulée")
        self.assertEqual(len(bdd.Tache.query.all()), 0)

        # two tasks and junk, abort
        bdd.Tache(id=3, timestamp=datetime.datetime.now(), commande="oz").add()
        bdd.Tache(id=5, timestamp=datetime.datetime.now(), commande="uz").add()
        ctx = mock_discord.get_ctx(cancel, "9", "az", "3", "5")
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent(["oz", "uz"], "aborted")
        self.assertEqual(len(bdd.Tache.query.all()), 2)

        # two tasks and junk, proceed
        ctx = mock_discord.get_ctx(cancel, "9", "az", "3", "5")
        with mock_discord.interact(("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent(["oz", "uz"], "annulée")
        self.assertEqual(len(bdd.Tache.query.all()), 0)



"""lg-rez / features / Synchronisation GSheets

Récupération et application des données des GSheets :
modifications décidées via le Tableau de bord et rôles

"""

import datetime
import time
import traceback

from discord import Embed
from discord.ext import commands
import sqlalchemy

from lgrez import config, bdd
from lgrez.blocs import tools, env, gsheets
from lgrez.bdd import (Joueur, Action, Role, Camp, BaseAction, BaseCiblage,
                       Statut, ActionTrigger)
from lgrez.features import gestion_actions


class TDBModif(gsheets.Modif):
    """Modification flag sur le Tableau de bord, à appliquer.

    Attributes:
        id (int): ID Discord du joueur concerné.
        col (str): Colonne de :class:`~bdd.Joueur` à modifier.
        val (Any): Nouvelle valeur.
        row (int): Numéro de la ligne (0 = ligne 1).
        column (int): Numéro de la colonne (0 = colonne A).

    Sous-classe de :class:`.gsheets.Modif`.
    """

    def __init__(self, id, col, val, row, column):
        """Initializes self."""
        super().__init__(row, column, val)
        self.id = id
        self.col = col

    def __repr__(self):
        """Returns repr(self)"""
        return f"<TDBModif id {self.id}: {self.col} = {self.val!r}>"

    def __eq__(self, other):
        """Returns self == other"""
        return (super().__eq__(other)
                and self.id == other.id
                and self.col == other.col)


class _ComparaisonResults():
    def __init__(self):
        self.n_id = 0
        self.upd = []
        self.add = []
        self.suppr = []
        self.sub_results = None

    @property
    def n_upd(self):
        return len(self.upd)

    @property
    def n_tot_attr_upd(self):
        return sum(len(up) for up in self.upd)

    @property
    def n_add(self):
        return len(self.add)

    @property
    def n_suppr(self):
        return len(self.suppr)

    def __iadd__(self, other):
        if isinstance(other, type(self)):
            self.n_id += other.n_id
            self.upd += other.upd
            self.add += other.add
            self.suppr += other.suppr
            if self.sub_results:
                self.sub_results += other.sub_results
            elif other.sub_results:
                self.sub_results = other.sub_results
            else:
                self.sub_results = None
            return self
        else:
            return NotImplemented

    def __add__(self, other):
        if isinstance(other, type(self)):
            new = type(self)()
            new += self
            new += other
            return new
        else:
            return NotImplemented

    @property
    def log(self):
        r = f"- {self.n_id} entrées identiques\n"
        if self.add:
            r += f"- Nouvelles entrées ({self.n_add}) : {self.add}\n"
        if self.upd:
            r += (f"- Entrées modifiées ({self.n_upd}, total attrs "
                  f"{self.n_tot_attr_upd}) : {self.upd}\n")
        if self.suppr:
            r += f"- Entrées supprimées ({self.n_suppr}) : {self.suppr}\n"
        return r

    @property
    def bilan(self):
        return (f"({self.n_id} entrées identiques - {self.n_upd} mises "
                f"à jour ({self.n_tot_attr_upd} attributs au total), "
                f"{self.n_add} ajouts, {self.n_suppr} suppressions")


def _compare_items(existants, new, table, cols, primary_key, bc_cols=None):
    """Utility function for !fillroles: compare table items and dicts"""
    res = _ComparaisonResults()
    if table == BaseAction:
        res.sub_results = _ComparaisonResults()

    # --- 5 : Pour chaque ligne : Mise à jour / Ajout
    for pk, args in new.items():
        pk = args[primary_key]
        if pk in existants:
            # Instance existante
            actual = existants[pk]
            attr_upd = []
            for col in cols.keys():
                if getattr(actual, col) != args[col]:
                    setattr(actual, col, args[col])
                    attr_upd.append(col)

            if table == BaseAction:
                # Many-to-many BaseAction <-> Rôle
                if set(actual.roles) != set(args["roles"]):
                    actual.roles = args["roles"]
                    attr_upd.append("roles")

                # BaseCiblages : on compare avec ceux existants !
                bcs = args["base_ciblages"]
                for bc_args in bcs:
                    bc_args["base_action"] = actual     # si création

                bc_existants = {bc.slug: bc for bc in actual.base_ciblages}
                new_bcs = {bc["slug"]: bc for bc in bcs}
                sub_res = _compare_items(bc_existants, new_bcs, BaseCiblage,
                                         cols=bc_cols, primary_key="slug")
                res.sub_results += sub_res

            if attr_upd:
                res.upd.append({actual: attr_upd})
            else:
                res.n_id += 1

        else:
            # Créer l'instance
            if table == BaseAction:     # BaseCiblages
                bcs = [BaseCiblage(**bc_args)
                       for bc_args in args["base_ciblages"]]
                args["base_ciblages"] = bcs
                res.sub_results.add.extend(bcs)
            inst = table(**args)
            inst.add()
            res.add.append(inst)

    # --- 6 : Drop anciennes instances
    for pk, act in existants.items():
        if pk not in new:
            config.session.delete(act)
            # pas .delete(), pour ne pas commit
            res.suppr.append(act.primary_key)

    return res


def transtype(value, cst):
    """Utilitaire : caste une donnée brute d'un GSheet selon sa colonne.

    Args:
        value (Any): valeur à transtyper.
        cst (:class:`sqlalchemy.schema.Column` | :class:`.bdd.base.TableMeta`\
            | :class:`sqlalchemy.orm.RelationshipProperty`): colonne, table ou
            relationship (many-to-one) associée.

    Types pris en charge dans le cas d'une colonne :

        - :class:`sqlalchemy.types.String` et dérivés
          (``Text``, ``Varchar``...)
        - :class:`sqlalchemy.types.Integer` et dérivés
          (``BigInteger``...)
        - :class:`sqlalchemy.types.Boolean`
        - :class:`sqlalchemy.types.Time`
        - :class:`sqlalchemy.types.Enum`

    Dans le cas d'une table ou d'une relation many-to-one vers une table,
    ``value`` est interprété comme la clé primaire de la table / de la table
    liée. Les valeurs interprétées ``None`` ne sont pas acceptées, même dans
    le cas d'une relation avec contrainte One-to-many faite *nullable*.

    Returns:
        L'objet Python correspondant au type de la colonne / table liée
        (:class:`str`, :class:`int`, :class:`bool`, :class:`datetime.time`,
        :class:`enum.Enum`, :class:`.bdd.base.TableBase`) ou ``None``

    Raises:
        ValueError: la conversion n'est pas possible (ou ``value`` est
            évaluée ``None`` et la colonne n'est pas *nullable*)
        TypeError: type de colonne non pris en charge.
    """
    if isinstance(cst, sqlalchemy.orm.RelationshipProperty):
        # Relationship
        table = cst.argument()    # -> cas suivant
        if not isinstance(value, table.primary_col.type.python_type):
            raise ValueError(
                f"Valeur '{value}' incorrecte pour la colonne '{cst.key}' "
                f"(one-to-many avec '{table.__name__}', de clé primaire "
                f"'{table.primary_col.name}', type '{table.primary_col.type}')"
            ) from None

        inst = table.query.get(value)
        if inst is None:
            raise ValueError(
                f"Valeur '{value}' incorrecte pour la colonne '{cst.key}': "
                f"instance de '{table.__name__}' correspondate non trouvée."
            )

        return inst

    elif isinstance(cst, bdd.base.TableMeta):
        # Table
        if not isinstance(value, cst.primary_col.type.python_type):
            raise ValueError(
                f"Valeur '{value}' incorrecte pour la table '{cst.__name__}', "
                f"de clé primaire '{cst.primary_col.name}', type "
                f"'{cst.primary_col.type}')"
            ) from None

        inst = cst.query.get(value)
        if inst is None:
            raise ValueError(
                f"Valeur '{value}' incorrecte pour la table '{cst.__name__}': "
                f"instance correspondate non trouvée."
            )

        return inst

    # Colonne propre
    try:
        if value in (None, '', 'None', 'none', 'Null', 'null'):
            if not cst.nullable:
                if cst.default is None:
                    raise ValueError
                else:
                    return cst.default.arg
            return None
        if isinstance(cst.type, sqlalchemy.Enum):
            enum = cst.type.python_type
            try:
                return enum[value]
            except KeyError:
                raise ValueError
        if isinstance(cst.type, sqlalchemy.String):
            return str(value)
        if isinstance(cst.type, sqlalchemy.Integer):
            return int(value)
        if isinstance(cst.type, sqlalchemy.Boolean):
            if (value in {True, 1}
                or (isinstance(value, str)
                    and value.lower() in {'true', 'vrai', '1'})):
                return True
            elif (value in {False, 0}
                  or (isinstance(value, str)
                      and value.lower() in {'false', 'faux', '0'})):
                return False
            else:
                raise ValueError
        if isinstance(cst.type, sqlalchemy.Time):       # hh:mm
            try:
                h, m, _ = value.split(':')
            except ValueError:
                h, m = value.split(':')
            return datetime.time(hour=int(h), minute=int(m))

        raise TypeError(f"Unhandled type for column '{cst.key}': '{cst.type}'")

    except (ValueError, TypeError):
        raise ValueError(
            f"Valeur '{value}' incorrecte pour la colonne '{cst.key}' "
            f"(type '{cst.type}'/{'NOT NULL' if not cst.nullable else ''})"
        ) from None


def get_sync():
    """Récupère les modifications en attente sur le TDB.

    Charge les données du Tableau de bord (variable d'environment
    ``LGREZ_TDB_SHEET_ID``), compare les informations qui y figurent
    avec celles de la base de données (:class:`.bdd.Joueur`).

    Supprime les joueurs en base absents du Tableau de bord, lève une
    erreur dans le cas inverse, n'applique aucune autre modification.

    Returns:
        list[.TDBModif]: La liste des modifications à apporter
    """
    # RÉCUPÉRATION INFOS GSHEET ET VÉRIFICATIONS

    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")
    workbook = gsheets.connect(SHEET_ID)
    sheet = workbook.worksheet(config.tdb_main_sheet)
    values = sheet.get_all_values()         # Liste de listes

    head = values[config.tdb_header_row - 1]
    # Ligne d'en-têtes (noms des colonnes), - 1 car indexé à 0

    id_index = gsheets.a_to_index(config.tdb_id_column)
    pk = head[id_index]
    if pk != Joueur.primary_col.key:
        raise ValueError(
            "Tableau de bord : la cellule "
            "`config.tdb_id_column` / `config.tdb_header_row` = "
            f"`{config.tdb_id_column}{config.tdb_header_row}` "
            f"vaut `{pk}` au lieu de la clé primaire de la table "
            f"`Joueur`, `{Joueur.primary_col.key}` !"
        )

    mstart, mstop = config.tdb_main_columns
    main_indexes = range(gsheets.a_to_index(mstart),
                         gsheets.a_to_index(mstop) + 1)
    # Indices des colonnes à remplir
    cols = {}
    for index in main_indexes:
        col = head[index]
        if col in Joueur.attrs:
            cols[col] = Joueur.attrs[col]
        else:
            raise ValueError(
                f"Tableau de bord : l'index de la zone principale "
                f"`{col}` n'est pas une colonne de la table `Joueur` !"
                " (voir `lgrez.config.main_indexes` / "
                "`lgrez.config.tdb_header_row`)"
            )

    tstart, tstop = config.tdb_tampon_columns
    tampon_indexes = range(gsheets.a_to_index(tstart),
                           gsheets.a_to_index(tstop) + 1)

    TDB_tampon_index = {}
    for index in tampon_indexes:
        col = head[index].partition("_")[2]
        if col in cols:
            TDB_tampon_index[col] = index
        else:
            raise ValueError(
                f"Tableau de bord : l'index de zone tampon `{head[index]}` "
                f"réfère à la colonne `{col}` (partie suivant le premier "
                f"underscore), qui n'est pas une colonne de la zone "
                "principale ! (voir `lgrez.config.tampon_indexes` / "
                "`lgrez.config.main_indexes`)"
            )

    # CONVERSION INFOS GSHEET EN PSEUDO-UTILISATEURS

    joueurs_TDB = []    # Joueurs tels qu'actuellement dans le TDB
    ids_TDB = []        # discord_ids des différents joueurs du TDB
    rows_TDB = {}       # Lignes ou sont les différents joueurs du TDB

    for i_row, row in enumerate(values):
        # On parcourt les lignes du TDB
        if i_row < config.tdb_header_row:
            # Ligne avant le header / le header (car décalage de 1)
            continue

        id_cell = row[id_index]
        if not id_cell.isdigit():
            # La cellule ne contient pas un ID ==> skip
            continue

        id = int(id_cell)
        # Construction dictionnaire correspondant à l'utilisateur
        joueur_TDB = {head[index]: transtype(row[index], cols[head[index]])
                      for index in main_indexes}
        joueur_TDB[pk] = id
        joueurs_TDB.append(joueur_TDB)
        ids_TDB.append(id)
        rows_TDB[id] = i_row

    # RÉCUPÉRATION UTILISATEURS BDD

    joueurs_BDD = {joueur.discord_id: joueur for joueur in Joueur.query.all()}

    # COMPARAISON

    for id, joueur in list(joueurs_BDD.items()):
        if id not in ids_TDB:
            # Joueur en base supprimé du TDB
            del joueurs_BDD[id]
            joueur.delete()

    modifs = []         # modifs à porter au TDB (liste de TDBModifs)
    for joueur_TDB in joueurs_TDB:              # Différences
        id = joueur_TDB[pk]

        try:
            joueur = joueurs_BDD[id]
        except KeyError:        # Joueur en base pas dans le TDB
            raise ValueError(f"Joueur `{joueur_TDB['nom']}` hors base : "
                             "vérifier processus d'inscription") from None

        for col in cols:
            if getattr(joueur, col) != joueur_TDB[col]:
                # Si <col> diffère entre TDB et cache,
                # on ajoute la modif (avec update du tampon)
                modifs.append(TDBModif(
                    id=id, col=col, val=joueur_TDB[col],
                    row=rows_TDB[id], column=TDB_tampon_index[col]
                ))

    return modifs


def validate_sync(modifs):
    """Valide des modificatons sur le Tableau de bord (case plus en rouge).

    Args:
        modifs (list[.TDBModif]): liste des modifications à apporter.

    Modifie sur le Tableau de bord (variable d'environment
    ``LGREZ_TDB_SHEET_ID``) et applique les modifications contenues
    dans ``modifs``.
    """
    SHEET_ID = env.load("LGREZ_TDB_SHEET_ID")
    workbook = gsheets.connect(SHEET_ID)    # Tableau de bord
    sheet = workbook.worksheet(config.tdb_main_sheet)

    gsheets.update(sheet, *modifs)


async def modif_joueur(joueur_id, modifs, silent=False):
    """Attribue les modifications demandées au joueur

    Args:
        joueur_id (int): id Discord du joueur concerné.
        modifs (list[.TDBModif]): liste des modifications à apporter.
        silent (bool): si ``True``, ne notifie pas le joueur des
            modifications.

    Returns:
        (list[.TDBModif], str): La liste des modifications appliquées
        et le changelog textuel associé (pour log global).

    Raises:
        ValueError: pas de joueur d'ID ``joueur_id`` en base

    Pour chaque modification dans ``modifs``, applique les conséquences
    adéquates (rôles, nouvelles actions, tâches planifiées...) et
    informe le joueur si ``silent`` vaut ``False``.
    """
    joueur = Joueur.query.get(joueur_id)
    if not joueur:
        raise ValueError(f"!sync : joueur d'ID {joueur_id} introuvable")

    member = joueur.member
    chan = joueur.private_chan

    changelog = (f"\n- {member.display_name} "
                 f"(@{member.name}#{member.discriminator}) :\n")
    notif = ""
    af = ":arrow_forward:"      # Flèche introduisant chaque modif

    if not modifs:
        changelog +=  f"    [NO MODIFS]\n"
        return [], changelog

    done = []
    for modif in modifs:
        changelog += f"    - {modif.col} : {modif.val}\n"

        if modif.col == "nom":                          # Renommage joueur
            await member.edit(nick=modif.val)
            await chan.edit(name=f"{config.private_chan_prefix}{modif.val}")
            if not silent:
                notif += (f"{af} Tu t'appelles maintenant "
                          f"{tools.bold(modif.val)}.\n")

        elif modif.col == "chambre" and not silent:     # Modification chambre
            notif += (f"{af} Tu habites maintenant "
                      f"en chambre {tools.bold(modif.val)}.\n")

        elif modif.col == "statut":
            if modif.val == Statut.vivant:              # Statut = vivant
                await member.add_roles(config.Role.joueur_en_vie)
                await member.remove_roles(config.Role.joueur_mort)
                if not silent:
                    notif += f"{af} Tu es maintenant en vie. EN VIE !!!\n"

            elif modif.val == Statut.mort:              # Statut = mort
                await member.add_roles(config.Role.joueur_mort)
                await member.remove_roles(config.Role.joueur_en_vie)
                if not silent:
                    notif += (f"{af} Tu es malheureusement décédé(e) :cry:\n"
                              "Ça arrive même aux meilleurs, en espérant "
                              "que ta mort ait été belle !\n")
                # Actions à la mort
                for action in joueur.actions_actives:
                    if action.base.trigger_debut == ActionTrigger.mort:
                        await gestion_actions.open_action(action)

            elif modif.val == Statut.MV:                # Statut = MV
                await member.add_roles(config.Role.joueur_en_vie)
                await member.remove_roles(config.Role.joueur_mort)
                if not silent:
                    notif += (
                        f"{af} Oh ! Tu viens d'être réduit(e) au statut de "
                        "mort-vivant... Un MJ viendra te voir très vite, "
                        "si ce n'est déjà fait, mais retient que la partie "
                        "n'est pas finie pour toi !\n"
                    )

            elif not silent:                            # Statut = autre
                notif += f"{af} Nouveau statut : {tools.bold(modif.val)} !\n"

        elif modif.col == "role":                       # Modification rôle
            new_role = modif.val
            for action in joueur.actions_actives:
                if action.base in joueur.role.base_actions:
                    # Suppression anciennes actions de rôle
                    gestion_actions.delete_action(action)

            for base in new_role.base_actions:
                # Ajout et création des tâches si trigger temporel
                action = Action(joueur=joueur, base=base, cooldown=0,
                                charges=base.base_charges)
                gestion_actions.add_action(action)

            if not silent:
                notif += (
                    f"{af} Ton nouveau rôle, si tu l'acceptes : "
                    f"{tools.bold(new_role.nom_complet)} !\nQue ce soit pour "
                    "un jour ou pour le reste de la partie, renseigne-toi en "
                    f"tapant {tools.code(f'!roles {new_role.nom}')}.\n"
                )

        elif modif.col == "camp" and not silent:    # Modification camp
            notif += (f"{af} Tu fais maintenant partie "
                      f"du camp « {tools.bold(modif.val.nom)} ».\n")

        elif modif.col == "votant_village" and not silent:
            if modif.val:                           # votant_village = True
                notif += (f"{af} Tu peux maintenant participer "
                          "aux votes du village !\n")
            else:                                   # votant_village = False
                notif += (f"{af} Tu ne peux maintenant plus participer "
                          "aux votes du village.\n")

        elif modif.col == "votant_loups" and not silent:
            if modif.val:                           # votant_loups = True
                notif += (f"{af} Tu peux maintenant participer "
                          "aux votes des loups ! Amuse-toi bien :wolf:\n")
            else:                                   # votant_loups = False
                notif += (f"{af} Tu ne peux maintenant plus participer "
                          "aux votes des loups.\n")

        elif modif.col == "role_actif" and not silent:
            if modif.val:                           # role_actif = True
                notif += (f"{af} Tu peux maintenant utiliser tes pouvoirs !\n")
            else:                                   # role_actif = False
                notif += (f"{af} Tu ne peux maintenant plus utiliser "
                          "aucun pouvoir.\n")

        setattr(joueur, modif.col, modif.val)
        # Dans tous les cas, on modifie en base
        # (après, pour pouvoir accéder aux vieux attribus plus haut)
        done.append(modif)

    if not silent:
        await chan.send(
            f":zap: {member.mention} Une action divine vient "
            f"de modifier ton existence ! :zap:\n\n{notif}\n"
            + tools.ital(":warning: Si tu penses qu'il y a erreur, "
                         "appelle un MJ au plus vite !")
        )

    return done, changelog


class Sync(commands.Cog):
    """Commandes de synchronisation des GSheets vers la BDD et les joueurs"""

    @commands.command()
    @tools.mjs_only
    async def sync(self, ctx, silent=False):
        """Récupère et applique les modifs du Tableau de bord (COMMANDE MJ)

        Args:
            silent: si spécifié (quelque soit sa valeur), les joueurs
                ne sont pas notifiés des modifications.

        Cette commande va récupérer les modifications en attente sur le
        Tableau de bord (lignes en rouge), modifer la BDD, et appliquer
        les modificatons dans Discord le cas échéant : renommage des
        utilisateurs, modification des rôles...
        """
        await ctx.send("Récupération des modifications...")
        async with ctx.typing():
            # Récupération de la liste des modifs
            modifs = get_sync()
            silent = bool(silent)
            changelog = f"Synchronisation TDB (silencieux = {silent}) :"

        if not modifs:
            await ctx.send("Pas de nouvelles modificatons.")
            return

        dic = {}        # Dicionnaire {ID joueur: modifs}
        for modif in modifs:
            if modif.id not in dic:
                dic[modif.id] = []
            dic[modif.id].append(modif)

        message = await ctx.send(
            f"{len(modifs)} modification(s) trouvée(s) "
            f"pour {len(dic)} joueur(s), go ? (silent = {silent})"
        )
        if not await tools.yes_no(message):
            await ctx.send("Mission aborted.")
            return

        # Go sync
        done = []
        async with ctx.typing():
            for joueur_id, modifs in dic.items():
                # Joueur dont au moins un attribut a été modifié
                try:
                    dn, cgl = await modif_joueur(int(joueur_id),
                                                 modifs, silent)
                except Exception:
                    # Erreur lors d'une des modifs
                    changelog += traceback.format_exc()
                    await ctx.send(
                        f"Erreur joueur {joueur_id}, passage au suivant "
                        "(voir logs pour les détails)"
                    )
                else:
                    # Pas d'erreur pour ce joueur, on enregistre
                    done.extend(dn)
                    changelog += cgl

            if done:
                # Au moins une modification a été appliquée
                config.session.commit()
                validate_sync(done)

            await tools.log(changelog, code=True)

        await ctx.send(
            f"Fait (voir {config.Channel.logs.mention} pour le détail)"
        )


    @commands.command()
    @tools.mjs_only
    async def fillroles(self, ctx):
        """Remplit les tables et #roles depuis le GSheet ad hoc (COMMANDE MJ)

        - Remplit les tables :class:`.bdd.Camp`, :class:`.bdd.Role`,
          :class:`.bdd.BaseAction` et :class:`.bdd.BaseCiblage` avec les
          informations du Google Sheets "Rôles et actions" (variable
          d'environnement ``LGREZ_ROLES_SHEET_ID``) ;
        - Vide le chan ``#roles`` puis le remplit avec les descriptifs
          de chaque rôle et camp.

        Utile à chaque début de saison / changement dans les rôles/actions.
        Met à jour les entrées déjà en base, créé les nouvelles,
        supprime celles obsolètes.
        """
        # ==== Mise à jour tables ===
        SHEET_ID = env.load("LGREZ_ROLES_SHEET_ID")
        workbook = gsheets.connect(SHEET_ID)    # Rôles et actions

        for table in [Camp, Role, BaseAction]:
            await ctx.send(
                f"Remplissage de la table {tools.code(table.__name__)}..."
            )
            # --- 1 : Récupération des valeurs
            async with ctx.typing():
                try:
                    sheet = workbook.worksheet(table.__tablename__)
                except gsheets.WorksheetNotFound:
                    raise ValueError(
                        f"!fillroles : feuille '{table.__tablename__}' non "
                        "trouvée dans le GSheet *Rôles et actions* "
                        "(`LGREZ_ROLES_SHEET_ID`)"
                    ) from None

                values = sheet.get_all_values()
                # Liste de liste des valeurs des cellules

            # --- 2 : Détermination colonnes à récupérer
            cols = table.columns        # "dictionnaire" nom -> colonne
            cols = {col: cols[col] for col in cols.keys()
                    if not col.startswith("_")}     # Colonnes publiques
            if table == Role:
                cols["camp"] = Role.attrs["camp"]
            elif table == BaseAction:
                # BaseCiblages : au bout de la feuille
                bc_cols = BaseCiblage.columns
                bc_cols = {col: bc_cols[col]
                           for col in bc_cols.keys()
                           if not col.startswith("_")}
                bc_cols_for_ith = []
                for i in range(config.max_ciblages_per_action):
                    prefix = f"c{i + 1}_"
                    bc_cols_for_ith.append({
                        # colonne -> nom dans la table pour le ièmme
                        col: f"{prefix}{col}" for col in bc_cols
                    })
            primary_key = table.primary_col.key

            # --- 3 : Localisation des colonnes (indices GSheet)
            cols_index = {}
            try:
                for key in cols.keys():
                    cols_index[key] = values[0].index(key)
                if table == BaseAction:
                    key = "roles"
                    roles_idx = values[0].index(key)
                    # BaseCiblages : au bout de la feuille
                    ciblages_idx = []
                    for bc_cols_names in bc_cols_for_ith:
                        idx = {}
                        for col, key in bc_cols_names.items():
                            idx[col] = values[0].index(key)
                        ciblages_idx.append(idx)
            except ValueError:
                raise ValueError(
                    f"!fillroles : colonne '{key}' non trouvée dans "
                    f"la feuille '{table.__tablename__}' du GSheet "
                    "*Rôles et actions* (`LGREZ_ROLES_SHEET_ID`)"
                ) from None

            # --- 4 : Constrution dictionnaires de comparaison
            existants = {item.primary_key: item for item in table.query.all()}
            new = {}
            for row in values[1:]:
                args = {key: transtype(row[cols_index[key]], col)
                        for key, col in cols.items()}

                if table == BaseAction:
                    # Many-to-many BaseAction <-> Rôle
                    roles = row[roles_idx].strip()
                    if roles.startswith("#"):
                        args["roles"] = []
                    else:
                        args["roles"] = [transtype(slug.strip(), Role)
                                         for slug in roles.split(",")
                                         if slug]
                    # BaseCiblages
                    new_bcs = []
                    for idx in ciblages_idx:
                        if row[idx["slug"]]:        # ciblage défini
                            bc_args = {key: transtype(row[idx[key]], col)
                                       for key, col in bc_cols.items()}
                            new_bcs.append(bc_args)
                    args["base_ciblages"] = new_bcs

                new[args[primary_key]] = args

            # --- 5 : Comparaison et MAJ
            res = _compare_items(
                existants, new, table, cols, primary_key,
                bc_cols=bc_cols if table == BaseAction else None
            )
            config.session.commit()

            await ctx.send(f"> Table {tools.code(table.__name__)} remplie ! "
                           + res.bilan)
            await tools.log(f"`!fillroles` > {table.__name__}:\n{res.log}")
            if table == BaseAction:
                sub_res = res.sub_results
                await ctx.send(f"> Table {tools.code('BaseCiblage')} remplie "
                               "simultanément " + sub_res.bilan)
                await tools.log(f"`!fillroles` > BaseCiblage:\n{sub_res.log}")

        # ==== Remplissage #rôles ===
        chan_roles = config.Channel.roles
        mess = await ctx.send(f"Purger et re-remplir {chan_roles.mention} ?")
        if not await tools.yes_no(mess):
            await ctx.send(f"Fini (voir {config.Channel.logs.mention}).")
            return

        await ctx.send(f"Vidage de {chan_roles.mention}...")
        async with ctx.typing():
            await chan_roles.purge(limit=1000)

        camps = Camp.query.filter_by(public=True).all()
        est = sum(len(camp.roles) + 2 for camp in camps) + 1
        await ctx.send(f"Remplissage... (temps estimé : {est} secondes)")

        t0 = time.time()
        await chan_roles.send("Voici la liste des rôles "
                              f"(voir aussi {tools.code('!roles')}) :")
        async with ctx.typing():
            shortcuts = []
            for camp in camps:
                if not camp.roles:
                    continue

                embed = Embed(title=f"Camp : {camp.nom}",
                              description=camp.description,
                              color=0x64b9e9)
                if (emoji := camp.discord_emoji_or_none):
                    embed.set_image(url=emoji.url)
                mess = await chan_roles.send(camp.nom, embed=embed)
                shortcuts.append(mess)

                for role in camp.roles:
                    if role.actif:
                        await chan_roles.send(embed=role.embed)

            for mess in shortcuts:
                await mess.reply("Accès rapide :             "
                                 "\N{UPWARDS BLACK ARROW}")

        rt = time.time() - t0
        await ctx.send(f"{chan_roles.mention} rempli ! (en {rt:.4} secondes)")



import unittest
from unittest import mock

import discord

from lgrez import config, bdd
from lgrez.features import annexe
from test import mock_discord, mock_bdd



class TestAnnexe(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.annexe commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = annexe.Annexe(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()


    async def test_roll(self):
        """Unit tests for !roll command."""
        # async def roll(self, ctx, *, XdY)
        roll = self.cog.roll

        samples = {
            "": False,
            "grzz": False,
            "1d6": True,
            "17d6": True,
            "17d6+3": True,
            "17d6-3": True,
            "17d6 + 3": True,
            "17d6 - 3": True,
            "17d6 + 3d15 + 2": True,
            "17d6-3d15-2": True,
            "17d0-3d15-2": False,
            "0d6-3d15-2": False,
            "17d6-3d15-0": True,
        }

        for sample, result in samples.items():
            ctx = mock_discord.get_ctx(roll, XdY=sample)
            if result:
                await ctx.invoke()
                ctx.assert_sent("=")
            else:
                with self.assertRaises(discord.ext.commands.UserInputError):
                    await ctx.invoke()


    async def test_coinflip(self):
        """Unit tests for !coinflip command."""
        # async def coinflip(self, ctx)
        coinflip = self.cog.coinflip

        ctx = mock_discord.get_ctx(coinflip)
        p, c = 0, 0
        for i in range(1000):
            await ctx.invoke()
            if ctx.send.call_args.args[0] == "Pile":
                p += 1
            elif ctx.send.call_args.args[0] == "Face":
                c += 1
            else:
                raise AssertionError("returned wtf thing")

        self.assertGreater(p, 400)
        self.assertGreater(c, 400)


    @unittest.SkipTest
    async def test_ping(self):
        """Unit tests for !ping command."""
        # async def ping(self, ctx)
        ping = self.cog.ping


    @mock_bdd.patch_db
    async def test_addhere(self):
        """Unit tests for !addhere command."""
        # async def addhere(self, ctx, *joueurs)
        addhere = self.cog.addhere

        mock_bdd.add_campsroles(10, 10)
        joueurs = [
            bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                       statut=bdd.Statut.vivant, _role_slug="role1",
                       votant_village=True, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                       chambre="ch2", statut=bdd.Statut.vivant,
                       votant_village=True, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                       statut=bdd.Statut.mort, _role_slug="role3",
                       _camp_slug="camp3", votant_village=True,
                       votant_loups=False, role_actif=True),
            bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4",
                       chambre="ch 4", statut=bdd.Statut.mort,
                       votant_village=False, votant_loups=False,
                       role_actif=True),
            bdd.Joueur(discord_id=5, chan_id_=51, nom="Joueur 5",
                       statut=bdd.Statut.MV, _camp_slug="camp3",
                       votant_village=False, votant_loups=False,
                       role_actif=False),
            bdd.Joueur(discord_id=6, chan_id_=61, nom="Joueur6",
                       statut=bdd.Statut.immortel, _role_slug="role3",
                       votant_village=False, votant_loups=True,
                       role_actif=False),
        ]
        bdd.Joueur.add(*joueurs)

        # cible = bat filtre
        ctx = mock_discord.get_ctx(addhere, "biz=oui")
        with self.assertRaises(discord.ext.commands.UserInputError) as cm:
            await ctx.invoke()
        self.assertIn("critère 'biz' incorrect", cm.exception.args[0])

        # cible = all filtres
        crits = {
            "discord_id=3": [2],
            "chan_id_=21": [1],
            "nom=Joueur4": [3],
            "nom=Joueur 5": [4],
            "chambre=ch2": [1],
            "chambre=ch 4": [3],
            "statut=vivant": [0, 1],
            "statut=mort": [2, 3],
            "statut=MV": [4],
            "statut=immortel": [5],
            "role=role1": [0],
            "role=role3": [2, 5],
            "camp=camp3": [2, 4],
            "votant_village=True": [0, 1, 2],
            "votant_village=1": [0, 1, 2],
            "votant_village=faux": [3, 4, 5],
            "votant_loups=vrai": [5],
            "votant_loups=0": [0, 1, 2, 3, 4],
            "role_actif=True": [2, 3],
            "role_actif=False": [0, 1, 4, 5],
        }
        for crit, ijs in crits.items():
            ctx = mock_discord.get_ctx(addhere, crit)
            with mock_discord.mock_members_and_chans(*joueurs):
                members = [joueur.member for joueur in joueurs]
                with mock_discord.interact(("yes_no", False)):
                    await ctx.invoke()
            ctx.assert_sent(*[joueurs[ij].nom for ij in ijs], "Fini")
            self.assertEqual(ctx.channel.set_permissions.call_count, len(ijs))
            ctx.channel.set_permissions.assert_has_calls([
                mock.call(member, read_messages=True, send_messages=True)
                for ij, member in enumerate(members)
                if ij in ijs
            ])
            ctx.channel.purge.assert_not_called()

        # cible = non-existing joueur, correct
        ctx = mock_discord.get_ctx(addhere, "gouzigouzi")
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "krr")),
            ("wait_for_message_here", mock_discord.message(ctx, "Joueur2")),
            ("yes_no", False),
        ):
            with mock_discord.mock_members_and_chans(*joueurs):
                members = [joueur.member for joueur in joueurs]
                await ctx.invoke()
        ctx.assert_sent("Aucune entrée trouvée", "Aucune entrée trouvée",
                        "Joueur2", "Fini")
        ctx.channel.set_permissions.assert_called_once_with(
            members[1], read_messages=True, send_messages=True
        )
        ctx.channel.purge.assert_not_called()

        # cible = existing joueur
        ctx = mock_discord.get_ctx(addhere, "Joueur 5")
        with mock_discord.mock_members_and_chans(*joueurs):
            members = [joueur.member for joueur in joueurs]
            with mock_discord.interact(("yes_no", False)):
                await ctx.invoke()
        ctx.assert_sent("Joueur 5", "Fini")
        ctx.channel.set_permissions.assert_called_once_with(
            members[4], read_messages=True, send_messages=True
        )
        ctx.channel.purge.assert_not_called()

        # cible = several joueurs, correct some
        ctx = mock_discord.get_ctx(addhere, "Joueur2", "kwdzz", "Joueur 5")
        with mock_discord.interact(
            ("wait_for_message_here", mock_discord.message(ctx, "krr")),
            ("wait_for_message_here", mock_discord.message(ctx, "Joueur4")),
            ("yes_no", False),
        ):
            with mock_discord.mock_members_and_chans(*joueurs):
                members = [joueur.member for joueur in joueurs]
                await ctx.invoke()
        ctx.assert_sent("Aucune entrée trouvée", "Aucune entrée trouvée",
                        "Joueur2", "Joueur4", "Joueur 5", "Fini")
        self.assertEqual(ctx.channel.set_permissions.call_count, 3)
        ctx.channel.set_permissions.assert_has_calls([
            mock.call(members[1], read_messages=True, send_messages=True),
            mock.call(members[3], read_messages=True, send_messages=True),
            mock.call(members[4], read_messages=True, send_messages=True),
        ])
        ctx.channel.purge.assert_not_called()

        # cible = existing joueur, purge
        ctx = mock_discord.get_ctx(addhere, "Joueur 5")
        with mock_discord.mock_members_and_chans(*joueurs):
            members = [joueur.member for joueur in joueurs]
            with mock_discord.interact(("yes_no", True)):
                await ctx.invoke()
        ctx.assert_sent("Joueur 5", "Fini")
        ctx.channel.set_permissions.assert_called_once_with(
            members[4], read_messages=True, send_messages=True
        )
        ctx.channel.purge.assert_called_once()


    async def test_purge(self):
        """Unit tests for !purge command."""
        # async def purge(self, ctx, N=None)
        purge = self.cog.purge

        # purge all, abort
        ctx = mock_discord.get_ctx(purge)
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent("Supprimer tous")
        ctx.channel.purge.assert_not_called()

        # purge all, proceed
        ctx = mock_discord.get_ctx(purge)
        with mock_discord.interact(("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent("Supprimer tous")
        ctx.channel.purge.assert_called_once_with(limit=None)

        # purge 15, abort
        ctx = mock_discord.get_ctx(purge, "15")
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent("Supprimer les 15")
        ctx.channel.purge.assert_not_called()

        # purge 15, proceed
        ctx = mock_discord.get_ctx(purge, "15")
        with mock_discord.interact(("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent("Supprimer les 15")
        ctx.channel.purge.assert_called_once_with(limit=17)


    @unittest.SkipTest
    async def test_akinator(self):
        """Unit tests for !akinator command."""
        # async def akinator(self, ctx)
        akinator = self.cog.akinator


    @unittest.SkipTest
    async def test_xkcd(self):
        """Unit tests for !xkcd command."""
        # async def xkcd(self, ctx, N)
        xkcd = self.cog.xkcd



"""lg-rez / features / Commandes de gestion des salons

Création, ajout, suppression de membres

"""

import asyncio
import functools
import datetime

from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools, one_command
from lgrez.bdd import Joueur, Boudoir, Bouderie
from lgrez.features.sync import transtype



def in_boudoir(callback):
    """Décorateur : commande utilisable dans un boudoir uniquement.

    Lors d'une invocation de la commande décorée hors d'un boudoir
    (enregistré dans :class:`.bdd.Boudoir`), affiche un message d'erreur.

    Ce décorateur n'est utilisable que sur une commande définie dans un
    Cog.
    """
    @functools.wraps(callback)
    async def new_callback(self, ctx, *args, **kwargs):
        try:
            Boudoir.from_channel(ctx.channel)
        except ValueError:
            await ctx.reply("Cette commande est invalide en dehors "
                            "d'un boudoir.")
        else:
            # if ctx.authors
            return await callback(self, ctx, *args, **kwargs)

    return new_callback


async def _invite(joueur, boudoir, invite_msg):
    """Invitation d'un joueur dans un boudoir (lancer comme tâche à part)"""
    pc = joueur.private_chan
    bc = boudoir.chan
    mess = await pc.send(
        f"{joueur.member.mention} {boudoir.gerant.nom} t'as invité(e) à "
        f"rejoindre son boudoir : « {boudoir.nom} » !\nAcceptes-tu ?"
    )

    if await tools.yes_no(mess):
        await boudoir.add_joueur(joueur)
        info = f"{joueur.nom} a rejoint le boudoir !"
        confirm = f"Tu as bien rejoint {bc.mention} !"
    else:
        info = f"{joueur.nom} a refusé l'invitation à rejoindre ce boudoir."
        confirm = "Invitation refusée."

    try:
        await invite_msg.reply(info)
    except discord.HTTPException:
        await bc.send(info)

    await mess.reply(confirm)



class GestionChans(commands.Cog):
    """Gestion des salons"""

    @commands.group(aliases=["boudoirs"])
    async def boudoir(self, ctx):
        """Gestion des boudoirs

        Les options relatives à un boudoir précis ne peuvent être
        exécutées que dans ce boudoir ; certaines sont réservées au
        gérant dudit boudoir.
        """
        if not ctx.invoked_subcommand:
            # Pas de sous-commande (correspondante)
            if ctx.subcommand_passed:
                # Tentative de subcommand
                raise commands.BadArgument(
                    f"Option '{ctx.subcommand_passed}' inconnue"
                )
            ctx.message.content = f"!help {ctx.invoked_with}"
            with one_command.bypass(ctx):
                await config.bot.process_commands(ctx.message)


    @boudoir.command(aliases=["liste"])
    @tools.joueurs_only
    @tools.private
    async def list(self, ctx):
        """Liste les boudoirs dans lesquels tu es"""
        joueur = Joueur.from_member(ctx.author)
        bouderies = joueur.bouderies

        if not bouderies:
            await ctx.reply(
                "Tu n'es dans aucun boudoir pour le moment.\n"
                f"{tools.code('!boudoir create')} pour en créer un."
            )
            return

        rep = "Tu es dans les boudoirs suivants :"
        for bouderie in bouderies:
            rep += f"\n - {bouderie.boudoir.chan.mention}"
            if bouderie.gerant:
                rep += " (gérant)"

        rep += "\n\nUtilise `!boudoir leave` dans un boudoir pour le quitter."

        await ctx.send(rep)


    @boudoir.command(aliases=["new", "creer", "créer"])
    @tools.joueurs_only
    @tools.private
    async def create(self, ctx, *, nom=None):
        """Crée un nouveau boudoir dont tu es gérant"""
        member = ctx.author
        joueur = Joueur.from_member(member)

        if not nom:
            await ctx.send("Comment veux-tu nommer ton boudoir ?\n"
                           + tools.ital("(`stop` pour annuler)"))
            mess = await tools.wait_for_message_here(ctx)
            nom = mess.content

        if len(nom) > 32:
            await ctx.send("Le nom des boudoirs est limité à 32 caractères.")
            return

        await ctx.send("Création du boudoir...")
        async with ctx.typing():
            now = datetime.datetime.now()
            categ = tools.channel(config.boudoirs_category_name)
            if len(categ.channels) >= 50:
                # Limitation Discord : 50 channels par catégorie
                ok = False
                N = 2
                while not ok:
                    nom_nouv_cat = f"{config.boudoirs_category_name} {N}"
                    categ_new = tools.channel(nom_nouv_cat,
                                              must_be_found=False)
                    if not categ_new:
                        categ = await categ.clone(name=nom_nouv_cat)
                        ok = True
                    elif len(categ_new.channels) < 50:
                        # Catégorie N pas pleine
                        categ = categ_new
                        ok = True
                    N += 1

            chan = await config.guild.create_text_channel(
                nom,
                topic=f"Boudoir crée le {now:%d/%m à %H:%M}. "
                      f"Gérant(e) : {joueur.nom}",
                category=categ,
            )

            boudoir = Boudoir(chan_id=chan.id, nom=nom, ts_created=now)
            boudoir.add()
            await boudoir.add_joueur(joueur, gerant=True)

            await chan.send(
                f"{member.mention}, voici ton boudoir ! "
                "Tu peux maintenant y inviter des gens avec la commande "
                "`!boudoir invite`."
            )

        await ctx.send(f"Ton boudoir a bien été créé : {chan.mention} !")
        await tools.log(f"Boudoir {chan.mention} créé par {joueur.nom}.")


    @boudoir.command(aliases=["add"])
    @tools.joueurs_only
    @in_boudoir
    async def invite(self, ctx, *, cible):
        """Invite un joueur à rejoindre ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut y inviter "
                            "de nouvelles personnes.")
            return

        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message="Qui souhaites-tu inviter ?"
        )
        if joueur in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} est déjà dans ce boudoir !")
            return

        mess = await ctx.send(f"Invitation envoyée à {joueur.nom}.")
        asyncio.create_task(_invite(joueur, boudoir, mess))
        # On envoie l'invitation en arrière-plan (libération du chan).


    @boudoir.command(aliases=["remove"])
    @tools.joueurs_only
    @in_boudoir
    async def expulse(self, ctx, *, cible):
        """Expulse un membre de ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut en expulser "
                            "des membres.")
            return

        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message="Qui souhaites-tu expulser ?"
        )
        if joueur not in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} n'est pas membre du boudoir !")
            return

        await boudoir.remove_joueur(joueur)
        await joueur.private_chan.send(f"Tu as été expulsé(e) du boudoir "
                                       f"« {boudoir.nom} ».")
        await ctx.send(f"{joueur.nom} a bien été expulsé de ce boudoir.")


    @boudoir.command(aliases=["quit"])
    @tools.joueurs_only
    @in_boudoir
    async def leave(self, ctx):
        """Quitte ce boudoir"""
        joueur = Joueur.from_member(ctx.author)
        boudoir = Boudoir.from_channel(ctx.channel)

        if boudoir.gerant == joueur:
            await ctx.send(
                "Tu ne peux pas quitter un boudoir que tu gères. "
                "Utilise `!boudoir transfer` pour passer les droits "
                "de gestion ou `!boudoir delete` pour le supprimer."
            )
            return

        mess = await ctx.reply(
            "Veux-tu vraiment quitter ce boudoir ? Tu ne "
            "pourras pas y retourner sans invitation."
        )
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        await boudoir.remove_joueur(joueur)
        await ctx.send(tools.ital(f"{joueur.nom} a quitté ce boudoir."))


    @boudoir.command(aliases=["transmit"])
    @tools.joueurs_only
    @in_boudoir
    async def transfer(self, ctx, cible=None):
        """Transfère les droits de gestion de ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut en transférer "
                            "les droits de gestion.")
            return

        joueur = await tools.boucle_query_joueur(
            ctx, cible=cible, message=("À qui souhaites-tu confier "
                                       "la gestion de ce boudoir ?")
        )
        if joueur not in boudoir.joueurs:
            await ctx.send(f"{joueur.nom} n'est pas membre de ce boudoir !")
            return

        mess = await ctx.reply(
            "Veux-tu vraiment transférer les droits de ce boudoir ? "
            "Tu ne pourras pas les récupérer par toi-même."
        )
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        bd_gerant = next(bd for bd in boudoir.bouderies if bd.joueur == gerant)
        bd_nouv = next(bd for bd in boudoir.bouderies if bd.joueur == joueur)

        bd_gerant.gerant = False
        bd_nouv.gerant = True
        bd_nouv.ts_promu = datetime.datetime.now()
        Bouderie.update()
        await boudoir.chan.edit(
            topic=f"Boudoir crée le {boudoir.ts_created:%d/%m à %H:%M}. "
                  f"Gérant(e) : {joueur.nom}"
        )

        await ctx.send(f"Boudoir transféré à {joueur.nom}.")


    @boudoir.command()
    @tools.joueurs_only
    @in_boudoir
    async def delete(self, ctx):
        """Supprime ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut le supprimer.")
            return

        mess = await ctx.reply("Veux-tu vraiment supprimer ce boudoir ? "
                               "Cette action est irréversible.")
        if not await tools.yes_no(mess):
            await ctx.send("Mission aborted.")
            return

        await ctx.send("Suppression...")
        for joueur in boudoir.joueurs:
            await boudoir.remove_joueur(joueur)
            await joueur.private_chan.send(
                f"Le boudoir « {boudoir.nom } » a été supprimé."
            )

        await boudoir.chan.edit(name=f"\N{CROSS MARK} {boudoir.nom}")
        await ctx.send(tools.ital(
            "[Tous les joueurs ont été exclus de ce boudoir ; "
            "le channel reste présent pour archive.]"
        ))


    @boudoir.command()
    @tools.joueurs_only
    @in_boudoir
    async def rename(self, ctx, *, nom):
        """Renomme ce boudoir"""
        boudoir = Boudoir.from_channel(ctx.channel)
        gerant = Joueur.from_member(ctx.author)
        if boudoir.gerant != gerant:
            await ctx.reply("Seul le gérant du boudoir peut le renommer.")
            return

        if not nom:
            await ctx.send("Comment veux-tu renommer ce boudoir ?\n"
                           + tools.ital("(`stop` pour annuler)"))
            mess = await tools.wait_for_message_here(ctx)
            nom = mess.content

        if len(nom) > 32:
            await ctx.send("Le nom des boudoirs est limité à 32 caractères.")
            return

        boudoir.nom = nom
        boudoir.update()
        await boudoir.chan.edit(name=nom)
        await ctx.send("Boudoir renommé avec succès.")


    @commands.command()
    @tools.mjs_only
    async def addhere(self, ctx, *joueurs):
        """Ajoute les membres au chan courant (COMMANDE MJ)

        Args:
            *joueurs: membres à ajouter, chacun entouré par des
                guillemets si nom + prénom

        Si ``*joueurs`` est un seul élément, il peut être de la forme
        ``<crit>=<filtre>`` tel que décrit dans l'aide de ``!send``.
        """
        ts_debut = ctx.message.created_at - datetime.timedelta(microseconds=1)

        if len(joueurs) == 1 and "=" in joueurs[0]:
            # Si critère : on remplace joueurs
            crit, _, filtre = joueurs[0].partition("=")
            crit = crit.strip()
            if crit in Joueur.attrs:
                col = Joueur.attrs[crit]
                arg = transtype(filtre.strip(), col)
                joueurs = Joueur.query.filter_by(**{crit: arg}).all()
            else:
                raise commands.UserInputError(f"critère '{crit}' incorrect")
        else:
            # Sinon, si noms / mentions
            joueurs = [await tools.boucle_query_joueur(ctx, cible)
                       for cible in joueurs]

        for joueur in joueurs:
            await ctx.channel.set_permissions(joueur.member,
                                              read_messages=True,
                                              send_messages=True)
            await ctx.send(f"{joueur.nom} ajouté")

        mess = await ctx.send("Fini, purge les messages ?")
        if await tools.yes_no(mess):
            await ctx.channel.purge(after=ts_debut)


    @commands.command()
    @tools.mjs_only
    async def purge(self, ctx, N=None):
        """Supprime tous les messages de ce chan (COMMANDE MJ)

        Args:
            N: nombre de messages à supprimer (défaut : tous)
        """
        if N:
            mess = await ctx.send(
                f"Supprimer les {N} messages les plus récents de ce chan ? "
                "(sans compter le `!purge` et ce message)"
            )
        else:
            mess = await ctx.send("Supprimer tous les messages de ce chan ?")

        if await tools.yes_no(mess):
            await ctx.channel.purge(limit=int(N) + 2 if N else None)
